import json
from typing import TYPE_CHECKING, Optional, Iterable, Any, Union

from boxsdk.util.text_enum import TextEnum
from .base_item import BaseItem
from ..exception import BoxAPIException
from .metadata import Metadata
from ..util.api_call_decorator import api_call
from ..pagination.marker_based_dict_collection import MarkerBasedDictCollection
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection

if TYPE_CHECKING:
    from boxsdk.object.watermark import Watermark
    from boxsdk.object.group import Group
    from boxsdk.object.user import User
    from boxsdk.object.collaboration import Collaboration
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class ClassificationType(TextEnum):
    """An enum of possible classification types"""
    PUBLIC = 'Public'
    INTERNAL = 'Internal'
    CONFIDENTIAL = 'Confidential'
    NONE = 'None'


class Item(BaseItem):
    """Box API endpoint for interacting with files and folders."""

    _classification_template_key = 'securityClassification-6VMVochwUWo'

    def _get_accelerator_upload_url(self, file_id: Optional[str] = None) -> Optional[str]:
        """
        Make an API call to get the Accelerator upload url for either upload a new file or updating an existing file.

        :param file_id:
            Box id of the file to be uploaded. Not required for new file uploads.
        :return:
            The Accelerator upload url or None if cannot get the Accelerator upload url.
        """
        if file_id:
            self.validate_item_id(file_id)
        endpoint = f'{file_id}/content' if file_id else 'content'
        url = f'{self._session.api_config.BASE_API_URL}/files/{endpoint}'
        try:
            response_json = self._session.options(
                url=url,
                expect_json_response=True,
            ).json()
            return response_json.get('upload_url', None)
        except BoxAPIException:
            return None

    def _preflight_check(
            self, size: int,
            name: str = None,
            file_id: str = None,
            parent_id: str = None
    ) -> Optional[str]:
        """
        Make an API call to check if certain file can be uploaded to Box or not.
        (https://developer.box.com/en/guides/uploads/check/)

        Returns an accelerator URL if available, which comes for free in the response.

        :param size:
            The size of the file to be uploaded in bytes. Specify 0 for unknown file sizes.
        :param name:
            The name of the file to be uploaded. This is optional if `file_id` is specified,
            but required for new file uploads.
        :param file_id:
            Box id of the file to be uploaded. Not required for new file uploads.
        :param parent_id:
            The ID of the parent folder. Required only for new file uploads.
        :return:
            The Accelerator upload url or None if cannot get the Accelerator upload url.
        :raises:
            :class:`BoxAPIException` when preflight check fails.
        """
        if file_id:
            self.validate_item_id(file_id)
        endpoint = f'{file_id}/content' if file_id else 'content'
        url = f'{self._session.api_config.BASE_API_URL}/files/{endpoint}'
        data = {'size': size}
        if name:
            data['name'] = name
        if parent_id:
            data['parent'] = {'id': parent_id}

        response_json = self._session.options(
            url=url,
            expect_json_response=True,
            data=json.dumps(data),
        ).json()
        return response_json.get('upload_url', None)

    @api_call
    def update_info(self, *, data: dict, etag: Optional[str] = None, **kwargs: Any) -> 'Item':
        """
        Baseclass override.
        :param data:
            The updated information about this object.
            Must be JSON serializable.
            Update the object attributes in data.keys(). The semantics of the
            values depends on the the type and attributes of the object being
            updated. For details on particular semantics, refer to the Box
            developer API documentation <https://developer.box.com/>.
        :param etag:
            If specified, instruct the Box API to perform the update only if
            the current version's etag matches.
        :return:
            The updated object.
            Return a new object of the same type, without modifying the original object passed as self.
            Construct the new object with all the default attributes that are returned from the endpoint.
        """
        # pylint:disable=arguments-differ
        self.validate_item_id(self._object_id)
        headers = {'If-Match': etag} if etag is not None else None
        return super().update_info(data=data, headers=headers, **kwargs)

    @api_call
    def get(self, *, fields: Iterable[str] = None, etag: Optional[str] = None, **kwargs) -> 'Item':
        """
        Base class override.

        :param fields:
            List of fields to request.
        :param etag:
            If specified, instruct the Box API to get the info only if the current version's etag doesn't match.
        :returns:
            Information about the file or folder.
        :raises: :class:`BoxAPIException` if the specified etag matches the latest version of the item.
        """
        # pylint:disable=arguments-differ,arguments-renamed
        self.validate_item_id(self._object_id)
        headers = {'If-None-Match': etag} if etag is not None else None
        return super().get(fields=fields, headers=headers, **kwargs)

    @api_call
    def remove_shared_link(self, *, etag: Optional[str] = None, **kwargs: Any) -> bool:
        """
        Baseclass override.

        :param etag:
            If specified, instruct the Box API to delete the link only if the current version's etag matches.
        :param kwargs:
            Used to fulfill the contract of overriden method
        :returns:
            Whether or not the update was successful.
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the item.
        """
        # pylint:disable=arguments-differ
        return super().remove_shared_link(etag=etag)

    @api_call
    def delete(self, *, params: dict = None, etag: Optional[str] = None, **kwargs) -> bool:
        """Delete the item.

        :param params:
            Additional parameters to send with the request.
        :param etag:
            If specified, instruct the Box API to delete the item only if the current version's etag matches.
        :returns:
            Whether or not the delete was successful.
        :raises: :class:`BoxAPIException` if the specified etag doesn't match the latest version of the item.
        """
        # pylint:disable=arguments-differ,arguments-renamed
        self.validate_item_id(self._object_id)
        headers = {'If-Match': etag} if etag is not None else None
        return super().delete(params=params, headers=headers, **kwargs)

    def metadata(self, scope: str = 'global', template: str = 'properties') -> Metadata:
        """
        Instantiate a :class:`Metadata` object associated with this item.

        :param scope:
            Scope of the metadata. Must be either 'global' or 'enterprise'.
        :param template:
            The name of the metadata template.
            See https://developer.box.com/en/reference/resources/metadata/ for more details.
        :return:
            A new metadata instance associated with this item.
        """
        self.validate_item_id(self._object_id)
        return Metadata(self._session, self, scope, template)

    def get_all_metadata(self) -> MarkerBasedDictCollection:
        """
        Get all metadata attached to the item.
        """
        self.validate_item_id(self._object_id)
        return MarkerBasedDictCollection(
            session=self._session,
            url=self.get_url('metadata'),
            limit=None,
            marker=None,
            return_full_pages=False,
        )

    @api_call
    def get_watermark(self) -> 'Watermark':
        """
        Return the watermark info for a Box file

        :return:
            Watermark object.
        """
        self.validate_item_id(self._object_id)
        url = self.get_url('watermark')
        box_response = self._session.get(url)
        response = box_response.json()
        return self.translator.get('watermark')(response['watermark'])

    @api_call
    def apply_watermark(self) -> 'Watermark':
        """
        Apply watermark on a Box file

        :return:
            Watermark object.
        """
        self.validate_item_id(self._object_id)
        url = self.get_url('watermark')
        body_attributes = {
            'watermark': {
                'imprint': 'default'
            }
        }
        box_response = self._session.put(url, data=json.dumps(body_attributes))
        response = box_response.json()
        return self.translator.get('watermark')(response['watermark'])

    @api_call
    def delete_watermark(self) -> bool:
        """
        Deletes the watermark info for a Box file

        :return:
            Whether or not the delete succeeded.
        """
        self.validate_item_id(self._object_id)
        url = self.get_url('watermark')
        box_response = self._session.delete(url, expect_json_response=False)
        return box_response.ok

    @api_call
    def collaborate(
            self,
            accessible_by: Union['User', 'Group'],
            role: str,
            can_view_path: Optional[bool] = None,
            notify: Optional[bool] = None,
            fields: Iterable[str] = None
    ) -> 'Collaboration':
        """Collaborate user or group onto a Box item.

        :param accessible_by:
            An object containing the collaborator.
        :param role:
            The permission level to grant the collaborator.
        :param can_view_path:
            Indicates whether the user can view the path of the item collaborated into.  This can only be set for
            collaborations on folders.
        :param notify:
            Determines if the collaborator should receive a notification for the collaboration.
        :param fields:
            List of fields to request.
        :return:
            The new collaboration
        """
        self.validate_item_id(self._object_id)
        url = self._session.get_url('collaborations')
        body = {
            'item': {
                'type': self.object_type,
                'id': self.object_id,
            },
            'accessible_by': {
                'type': accessible_by.object_type,
                'id': accessible_by.object_id,
            },
            'role': role,
        }
        if can_view_path is not None:
            body['can_view_path'] = can_view_path
        params = {}
        if fields is not None:
            params['fields'] = ','.join(fields)
        if notify is not None:
            params['notify'] = notify
        response = self._session.post(url, data=json.dumps(body), params=params).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def collaborate_with_login(
            self,
            login: str,
            role: str,
            can_view_path: Optional[bool] = None,
            notify: Optional[bool] = None,
            fields: Iterable[str] = None
    ) -> 'Collaboration':
        """Collaborate user onto a Box item with the user login.

        :param login:
            The email address of the person to grant access to.
        :param role:
            The permission level to grant the collaborator.
        :param can_view_path:
            Indicates whether the user can view the path of the folder collaborated into.
        :param notify:
            Determines if the collaborator should receive a notification for the collaboration.
        :param fields:
            List of fields to request.
        :return:
            The new collaboration with the user login
        """
        self.validate_item_id(self._object_id)
        url = self._session.get_url('collaborations')
        body = {
            'item': {
                'type': self.object_type,
                'id': self.object_id,
            },
            'accessible_by': {
                'type': 'user',
                'login': login,
            },
            'role': role,
        }
        if can_view_path is not None:
            body['can_view_path'] = can_view_path
        params = {}
        if fields is not None:
            params['fields'] = ','.join(fields)
        if notify is not None:
            params['notify'] = notify
        response = self._session.post(url, data=json.dumps(body), params=params).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_collaborations(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the collaboration.

        :param limit:
            The maximum number of items to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start returning items from when using marker-based paging.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the collaboration.
        """
        self.validate_item_id(self._object_id)
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('collaborations'),
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    def add_classification(self, classification: str) -> str:
        """
        Applies metadata classification for the specified :class:`File` or :class:`Folder` object.

        :param classification:
            The classification to add to the :class:`File` or :class:`Folder`
        :return:
            The classification added to the :class:`File` or :class:`Folder.
        """
        classification_metadata = {
            'Box__Security__Classification__Key': classification,
        }
        metadata_classification = self.metadata(
            scope='enterprise',
            template=self._classification_template_key
        ).create(classification_metadata)
        return metadata_classification['Box__Security__Classification__Key']

    def update_classification(self, classification: str) -> str:
        """
        Updates metadata classification for the specified :class:`File` or :class:`Folder` object.

        :param classification:
            The classification to add to the :class:`File` or :class:`Folder`
        :return:
            The classification updated on the :class:`File` or :class:`Folder.
        """
        classification_metadata = self.metadata('enterprise', self._classification_template_key)
        updates = classification_metadata.start_update()
        updates.add('/Box__Security__Classification__Key', classification)
        metadata_classification = classification_metadata.update(updates)
        return metadata_classification['Box__Security__Classification__Key']

    def set_classification(self, classification: str) -> str:
        """
        Attempts to add a metadata classification to a :class:`File` or :class:`Folder`, if classification exists, then
        do update.

        :param classification:
            The classification to add to the :class:`File` or :class:`Folder`
        :return:
            The classification set on the :class:`File` or :class:`Folder.
        """
        classification_metadata = {
            'Box__Security__Classification__Key': classification,
        }
        return self.metadata(
            scope='enterprise',
            template=self._classification_template_key
        ).set(metadata=classification_metadata)['Box__Security__Classification__Key']

    def get_classification(self) -> Optional[str]:
        """
        Retrieves the classification specified for the :class:`File` or :class:`Folder`

        :return:
            The classification on the :class:`File` or :class:`Folder.
        """
        try:
            classification = self.metadata('enterprise', self._classification_template_key).get()
        except BoxAPIException as err:
            if err.status == 404 and err.code == "instance_not_found":
                return None
            raise
        return classification.get('Box__Security__Classification__Key', None)

    def remove_classification(self) -> bool:
        """
        Removes a metadata classification from a :class:`File` or :class:`Folder`.

        :returns:
            Whether or not the delete was successful.
        """
        return self.metadata('enterprise', self._classification_template_key).delete()



from .base_object import BaseObject


class CollaborationAllowlistExemptTarget(BaseObject):
    """Represents a user who is exempted from the collaboration allowlist."""

    _item_type = 'collaboration_whitelist_exempt_target'



from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from boxsdk.session.box_response import BoxResponse
    from boxsdk import NetworkResponse


def is_json_response(network_response: Union['BoxResponse', 'NetworkResponse']) -> bool:
    """Return whether or not the network response content is json.

    :param network_response:
        The response from the Box API.
    """
    try:
        network_response.json()
        return True
    except ValueError:
        return False



from datetime import datetime

import pytest
from boxsdk.pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection
from boxsdk.pagination.marker_based_object_collection import MarkerBasedObjectCollection

from test.integration_new import util
from test.integration_new import CLIENT
from test.integration_new.context_managers.box_test_file import BoxTestFile
from test.integration_new.context_managers.box_test_folder import BoxTestFolder

FILE_TESTS_DIRECTORY_NAME = 'file-integration-tests'


@pytest.fixture(scope="module", autouse=True)
def parent_folder():
    with BoxTestFolder(name=f'{FILE_TESTS_DIRECTORY_NAME} {datetime.now()}') as folder:
        yield folder


def test_trash_get_items(parent_folder, small_file_path):
    name = f'{util.random_name()}.pdf'
    test_file = parent_folder.upload(file_path=small_file_path, file_name=name)
    test_file.delete()
    try:
        trash_items = CLIENT.trash().get_items()
        assert test_file.id in [item.id for item in trash_items]
    finally:
        CLIENT.trash().permanently_delete_item(test_file)


def test_trash_restore_item(parent_folder, small_file_path):
    with BoxTestFile(parent_folder=parent_folder, file_path=small_file_path) as test_file:
        test_file.delete()
        trash_items = CLIENT.trash().get_items()
        assert test_file.id in [item.id for item in trash_items]
        CLIENT.trash().restore_item(test_file)
        folder_items = parent_folder.get_items()
        assert test_file.id in [item.id for item in folder_items]


def test_trash_get_items_with_offset(parent_folder, small_file_path):
    name = f'{util.random_name()}.pdf'
    test_file = parent_folder.upload(file_path=small_file_path, file_name=name)
    test_file.delete()
    try:
        trash_items = CLIENT.trash().get_items()
        assert isinstance(trash_items, LimitOffsetBasedObjectCollection)
        assert test_file.id in [item.id for item in trash_items]
    finally:
        CLIENT.trash().permanently_delete_item(test_file)


def test_trash_get_items_with_marker(parent_folder, small_file_path):
    name = f'{util.random_name()}.pdf'
    test_file = parent_folder.upload(file_path=small_file_path, file_name=name)
    test_file.delete()
    try:
        trash_items = CLIENT.trash().get_items(limit=100, use_marker=True)
        assert isinstance(trash_items, MarkerBasedObjectCollection)
        assert test_file.id in [item.id for item in trash_items]
    finally:
        CLIENT.trash().permanently_delete_item(test_file)



import json
import pytest


@pytest.fixture(params=('file', 'folder', 'web_link'))
def test_base_item_and_response(
        test_file, test_folder, test_web_link, mock_file_response, mock_folder_response, mock_web_link_response, request):
    if request.param == 'file':
        return test_file, mock_file_response
    if request.param == 'folder':
        return test_folder, mock_folder_response

    return test_web_link, mock_web_link_response


@pytest.fixture(params=('empty', 'same', 'other'))
def test_collections_for_addition(mock_collection_id, request):
    """Fixture returning a tuple of the expected collections values before and after addition"""
    other_collection_id = mock_collection_id + '2'
    if request.param == 'empty':
        return [], [{'id': mock_collection_id}]
    if request.param == 'same':
        # Adding a second instance of the same collection is handled correctly by the API,
        # so for simplicity we do not check for an existing copy of the collection and just append
        return [{'id': mock_collection_id}], [{'id': mock_collection_id}, {'id': mock_collection_id}]
    if request.param == 'other':
        return [{'id': other_collection_id}], [{'id': other_collection_id}, {'id': mock_collection_id}]

    raise NotImplementedError(f"Forgot to implement {request.param}")


@pytest.fixture(params=('empty', 'only_removed', 'only_other', 'other_and_removed'))
def test_collections_for_removal(mock_collection_id, request):
    """Fixture returning a tuple of the expected collections values before and after removal"""
    other_collection_id = mock_collection_id + '2'
    if request.param == 'empty':
        return [], []
    if request.param == 'only_removed':
        return [{'id': mock_collection_id}], []
    if request.param == 'only_other':
        return [{'id': other_collection_id}], [{'id': other_collection_id}]
    if request.param == 'other_and_removed':
        return [{'id': mock_collection_id}, {'id': other_collection_id}], [{'id': other_collection_id}]

    raise NotImplementedError(f"Forgot to implement {request.param}")


@pytest.mark.parametrize('params, expected_data', [
    ({}, {}),
    ({'name': 'New name.pdf'}, {'name': 'New name.pdf'})
])
def test_copy_base_item(test_base_item_and_response, mock_box_session, test_folder, mock_object_id, params, expected_data):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_base_item_and_response
    expected_url = test_item.get_url('copy')
    expected_body = {
        'parent': {'id': mock_object_id},
    }
    expected_body.update(expected_data)
    mock_box_session.post.return_value = mock_item_response
    copy_response = test_item.copy(parent_folder=test_folder, **params)
    mock_box_session.post.assert_called_once_with(expected_url, data=json.dumps(expected_body))
    assert isinstance(copy_response, test_item.__class__)


@pytest.mark.parametrize(
    'params, expected_data', [
        ({}, {}),
        ({'name': 'New name.pdf'}, {'name': 'New name.pdf'})
    ]
)
def test_move_base_item(test_base_item_and_response, mock_box_session, test_folder, mock_object_id, params, expected_data):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_base_item_and_response
    expected_url = test_item.get_url()
    expected_body = {
        'parent': {'id': mock_object_id},
    }
    expected_body.update(expected_data)
    mock_box_session.put.return_value = mock_item_response
    move_response = test_item.move(test_folder, **params)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(expected_body), params=None, headers=None)
    assert isinstance(move_response, test_item.__class__)


def test_rename_base_item(test_base_item_and_response, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_base_item_and_response
    expected_url = test_item.get_url()
    mock_box_session.put.return_value = mock_item_response
    rename_response = test_item.rename('new name')
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps({'name': 'new name'}), params=None, headers=None)
    assert isinstance(rename_response, test_item.__class__)


def test_add_to_collection(test_base_item_and_response, mock_box_session, mock_collection, test_collections_for_addition):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_base_item_and_response
    current_collections, expected_collections = test_collections_for_addition
    expected_url = test_item.get_url()
    expected_params = {'fields': 'collections'}
    expected_data = {
        'collections': expected_collections
    }
    mock_response = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'collections': current_collections,
    }
    mock_box_session.get.return_value.json.return_value = mock_response
    mock_box_session.put.return_value = mock_item_response

    test_item.add_to_collection(mock_collection)

    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=expected_params)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(expected_data), headers=None, params=None)


def test_remove_from_collection(test_base_item_and_response, mock_box_session, mock_collection, test_collections_for_removal):
    # pylint:disable=redefined-outer-name, protected-access
    test_item, mock_item_response = test_base_item_and_response
    current_collections, expected_collections = test_collections_for_removal
    expected_url = test_item.get_url()
    expected_params = {'fields': 'collections'}
    expected_data = {
        'collections': expected_collections
    }
    mock_response = {
        'type': test_item.object_type,
        'id': test_item.object_id,
        'collections': current_collections,
    }
    mock_box_session.get.return_value.json.return_value = mock_response
    mock_box_session.put.return_value = mock_item_response

    test_item.remove_from_collection(mock_collection)

    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=expected_params)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(expected_data), headers=None, params=None)



from typing import Any

from .base_object import BaseObject
from ..util.api_call_decorator import api_call


class SignRequest(BaseObject):
    """
    Represents a Sign Request used by Box Sign
    Sign Requests are used to request e-signatures on documents from signers.
    A Sign Request can refer to one or more Box Files and can be sent to one or more Box Sign Request Signers.
    """
    _item_type = 'sign-request'

    def get_url(self, *args: Any) -> str:
        """
        Returns the url for this sign request.
        """
        return self._session.get_url('sign_requests', self._object_id, *args)

    @api_call
    def cancel(self) -> 'SignRequest':
        """
        Cancels a sign request if it has not yet been signed or declined.
        Any outstanding signers will no longer be able to sign the document.

        :returns:
            The cancelled SignRequest object.
        """
        url = self.get_url('cancel')
        response = self._session.post(url).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def resend(self) -> bool:
        """
        Attempts to resend a Sign Request to all signers that have not signed yet.
        There is a 10 minute cooling-off period between each resend request.

        :returns:
            Whether the operation succeeded.
        """
        url = self.get_url('resend')
        response = self._session.post(url, skip_retry_codes={202}, expect_json_response=False)
        return response.ok



import json
from typing import Iterable, TYPE_CHECKING, Optional

from boxsdk.pagination.marker_based_object_collection import MarkerBasedObjectCollection

from .base_endpoint import BaseEndpoint

from ..pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection
from ..util.api_call_decorator import api_call

if TYPE_CHECKING:
    from boxsdk.object.base_item import BaseItem
    from boxsdk.object.folder import Folder
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class Trash(BaseEndpoint):
    """Box API endpoint for performing trash related actions in Box."""

    @api_call
    def get_item(self, item: 'BaseItem', fields: Iterable[str] = None) -> 'BaseItem':
        """
        Get item from trash.

        :param item:
            The :class:`BaseItem` object to get info on.
        :param fields:
            List of fields to request
        :returns:
            Information for a trashed :class:`BaseItem` object.
        """
        url = item.get_url('trash')
        params = {}
        if fields:
            params['fields'] = ','.join(fields)
        box_response = self._session.get(url, params=params)
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def restore_item(
            self,
            item: 'BaseItem',
            name: Optional[str] = None,
            parent_folder: Optional['Folder'] = None,
            fields: Iterable[str] = None
    ) -> 'BaseItem':
        """
        Restores an item from the trash. Could be files, folders, or weblinks.

        :param item:
            The :class:`BaseItem` object to restore from trash.
        :param name:
            The new name for this item. Only used if the item can't be restored due to name conflict.
        :param parent_folder:
            The new parent folder. Only used if the previous parent folder no longer exists.
        :param fields:
            List of fields to request
        :returns:
            A restored :class:`BaseItem`.
        """
        url = item.get_url()
        body = {}
        if name is not None:
            body['name'] = name
        if parent_folder is not None:
            body['parent'] = {'id': parent_folder.object_id}
        params = {}
        if fields:
            params['fields'] = ','.join(fields)
        box_response = self._session.post(
            url, data=json.dumps(body), params=params)
        response = box_response.json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def permanently_delete_item(self, item: 'BaseItem') -> bool:
        """
        Permanently delete an item that is in the trash. The item will no longer exist in Box.

        :param item:
            The :class:`BaseItem` to delete from trash.
        :returns:
            Whether or not the delete was successful.
        """
        url = item.get_url('trash')
        box_response = self._session.delete(url, expect_json_response=False)
        return box_response.ok

    @api_call
    def get_items(
        self,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        fields: Iterable[str] = None,
        marker: Optional[str] = None,
        use_marker: bool = False,
        sort: Optional[str] = None,
        direction: Optional[str] = None
    ) -> 'BoxObjectCollection':
        """
        Using limit-offset paging, get the files, folders and web links that are in the user's trash.

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param offset:
            The offset of the item at which to begin the response.
        :param fields:
            List of fields to request.
        :param marker:
            The marker at which to begin the response.
        :param use_marker:
            Whether or not to use marker-based paging.
        :param sort:
            The field to sort by. Can be 'id', 'name', 'date' or 'size'.
        :param direction:
            The direction to sort. Can be 'ASC' or 'DESC'.
        :returns:
            An iterator of the entries in the trash
        """
        additional_params = {}
        if limit is not None:
            additional_params['limit'] = limit
        if direction:
            additional_params['direction'] = direction
        if sort:
            additional_params['sort'] = sort

        if use_marker:
            additional_params['usemarker'] = True
            return MarkerBasedObjectCollection(
                url=self._session.get_url('folders', 'trash', 'items'),
                session=self._session,
                limit=limit,
                marker=marker,
                fields=fields,
                additional_params=additional_params,
                return_full_pages=False,
            )

        return LimitOffsetBasedObjectCollection(
            session=self._session,
            url=self._session.get_url('folders', 'trash', 'items'),
            limit=limit,
            offset=offset,
            fields=fields,
            return_full_pages=False,
            additional_params=additional_params,
        )



import pytest

from boxsdk.object.base_api_json_object import BaseAPIJSONObject
from boxsdk.object.base_object import BaseObject
from boxsdk.object.folder import Folder


@pytest.fixture(params=[{'foo': 'bar'}, {'a': {'b': 'c'}}])
def response(request):
    return request.param


@pytest.fixture()
def base_api_json_object(response):
    dictionary_response = response
    return dictionary_response, BaseAPIJSONObject(dictionary_response)


def test_getitem(base_api_json_object):
    dictionary_response, test_object = base_api_json_object
    assert isinstance(test_object, BaseAPIJSONObject)
    for key in dictionary_response:
        assert test_object[key] == dictionary_response[key]


def test_contains(base_api_json_object):
    dictionary_response, test_object = base_api_json_object
    for key in dictionary_response:
        assert key in test_object
    assert 'some_key_that_doesnt_exist' not in test_object


def test_iter(base_api_json_object):
    dictionary_response, test_object = base_api_json_object
    all_test_object_keys = list(test_object)
    all_dictionary_response_keys = list(dictionary_response)
    assert set(all_test_object_keys) == set(all_dictionary_response_keys)


def test_meta_registers_new_item_type_in_default_translator(default_translator, original_default_translator):
    item_type = 'ƒøø'

    class Foo(BaseAPIJSONObject):
        _item_type = item_type

    assert default_translator.get(item_type) is Foo
    assert (set(default_translator) - set(original_default_translator)) == {item_type}


@pytest.mark.parametrize('subclass', [BaseAPIJSONObject, BaseObject, Folder])
def test_meta_does_not_register_new_subclass_in_default_translator_if_item_type_is_not_defined_in_namespace(
        subclass,
        default_translator,
        original_default_translator,
):

    class Foo(subclass):
        pass

    assert Foo not in default_translator.values()
    assert default_translator == original_default_translator


def test_meta_overrides_registration_if_subclass_redefines_item_type(default_translator, original_default_translator):

    class FolderSubclass(Folder):
        _item_type = 'folder'

    assert default_translator.get('folder') is FolderSubclass
    assert set(default_translator.keys()) == set(original_default_translator.keys())



from datetime import datetime
from typing import Optional

from boxsdk.auth.server_auth import ServerAuth


class CCGAuth(ServerAuth):
    _GRANT_TYPE = 'client_credentials'

    def _fetch_access_token(self, subject_id: str, subject_type: str, now_time: Optional[datetime] = None) -> str:
        data = {
            'grant_type': self._GRANT_TYPE,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'box_subject_id': subject_id,
            'box_subject_type': subject_type,
        }

        return self.send_token_request(data, access_token=None, expect_refresh_token=False)[0]



from .auth import JWTAuth, OAuth2, CCGAuth
from .client import *  # pylint:disable=wildcard-import,redefined-builtin
from .exception import *  # pylint:disable=wildcard-import
from .object import *  # pylint:disable=wildcard-import,redefined-builtin
from .util.log import setup_logging
from .version import __version__



from typing import Any, Optional

from test.integration_new import util
from test.integration_new import CLIENT
from boxsdk.object.file import File
from boxsdk.object.folder import Folder


class BoxTestFile:

    def __init__(self, *, file_path: str, name: str = None, parent_folder: Optional[Folder] = None):
        if name is None:
            extention = file_path.split('.')[-1]
            name = f'{util.random_name()}.{extention}'
        if parent_folder is None:
            parent_folder = CLIENT.folder('0')
        self._file: File = parent_folder.upload(file_path=file_path, file_name=name)

    def __enter__(self) -> File:
        return self._file

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        util.permanently_delete(self._file)



import json
import pytest

from boxsdk.config import API
from boxsdk.object.enterprise import Enterprise
from boxsdk.object.folder import Folder
from boxsdk.object.metadata_cascade_policy import MetadataCascadePolicy, CascadePolicyConflictResolution


@pytest.fixture()
def test_cascade_policy(mock_box_session):
    return MetadataCascadePolicy(mock_box_session, 'test_cascade_policy_id')


def test_get(test_cascade_policy, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_cascade_policies/{test_cascade_policy.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'id': '84113349-794d-445c-b93c-d8481b223434',
        'type': 'metadata_cascade_policy',
        'owner_enterprise': {
            'type': 'enterprise',
            'id': '11111',
        },
        'parent': {
            'type': 'folder',
            'id': '22222',
        },
        'scope': 'enterprise_11111',
        'templateKey': 'testTemplate',
    }

    cascade_policy = test_cascade_policy.get()

    mock_box_session.get.assert_called_once_with(expected_url, params=None, headers=None)
    assert isinstance(cascade_policy, MetadataCascadePolicy)
    enterprise = cascade_policy.owner_enterprise
    assert isinstance(enterprise, Enterprise)
    assert enterprise.object_id == '11111'
    folder = cascade_policy.parent
    assert isinstance(folder, Folder)
    assert folder.object_id == '22222'
    assert cascade_policy.scope == 'enterprise_11111'
    assert cascade_policy.templateKey == 'testTemplate'


def test_delete(test_cascade_policy, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_cascade_policies/{test_cascade_policy.object_id}'
    mock_box_session.delete.return_value.ok = True

    result = test_cascade_policy.delete()

    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, headers=None, params={})
    assert result is True


def test_force_apply(test_cascade_policy, mock_box_session):
    expected_url = f'{API.BASE_API_URL}/metadata_cascade_policies/{test_cascade_policy.object_id}/apply'
    expected_body = {
        'conflict_resolution': 'overwrite',
    }
    mock_box_session.post.return_value.ok = True

    result = test_cascade_policy.force_apply(CascadePolicyConflictResolution.OVERWRITE)

    mock_box_session.post.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_body),
        expect_json_response=False
    )
    assert result is True



import json

from boxsdk.config import API
from boxsdk.object.web_link import WebLink
from boxsdk.util.datetime_formatter import normalize_date_to_rfc3339_format
from boxsdk.util.default_arg_value import SDK_VALUE_NOT_SET


def test_get(mock_box_session, test_web_link):
    # pylint:disable=redefined-outer-name, protected-access
    web_link_id = test_web_link.object_id
    expected_url = f'{API.BASE_API_URL}/web_links/{web_link_id}'
    mock_web_link = {
        'type': 'web_link',
        'url': 'https://test/com',
        'id': 1234,
        'created_at': '2015-05-07T15:00:01-07:00',
    }
    mock_box_session.get.return_value.json.return_value = mock_web_link
    web_link = test_web_link.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert web_link.id == mock_web_link['id']
    assert web_link.type == mock_web_link['type']
    assert web_link.url == mock_web_link['url']
    assert isinstance(web_link, WebLink)


def test_update(mock_box_session, test_web_link):
    # pylint:disable=redefined-outer-name, protected-access
    web_link_id = test_web_link.object_id
    expected_url = f'{API.BASE_API_URL}/web_links/{web_link_id}'
    mock_web_link = {
        'type': 'web_link',
        'url': 'https://newtest.com',
        'id': 1234,
        'created_at': '2015-05-07T15:00:01-07:00',
    }
    data = {
        'url': 'https://newtest.com',
    }
    mock_box_session.put.return_value.json.return_value = mock_web_link
    web_link = test_web_link.update_info(data=data)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert web_link.id == mock_web_link['id']
    assert web_link.type == mock_web_link['type']
    assert web_link.url == mock_web_link['url']
    assert isinstance(web_link, WebLink)


def test_delete(mock_box_session, test_web_link):
    web_link_id = test_web_link.object_id
    expected_url = f'{API.BASE_API_URL}/web_links/{web_link_id}'
    test_web_link.delete()
    mock_box_session.delete.assert_called_once_with(expected_url, expect_json_response=False, headers=None, params={})


def test_get_shared_link(
        test_web_link,
        mock_box_session,
        shared_link_access,
        shared_link_unshared_at,
        shared_link_password,
        shared_link_vanity_name,
        test_url,
):
    # pylint:disable=redefined-outer-name, protected-access
    expected_url = test_web_link.get_url()
    mock_box_session.put.return_value.json.return_value = {
        'type': test_web_link.object_type,
        'id': test_web_link.object_id,
        'shared_link': {
            'url': test_url,
        },
    }
    expected_data = {'shared_link': {}}
    if shared_link_access is not None:
        expected_data['shared_link']['access'] = shared_link_access
    if shared_link_unshared_at is not SDK_VALUE_NOT_SET:
        expected_data['shared_link']['unshared_at'] = normalize_date_to_rfc3339_format(shared_link_unshared_at)
    if shared_link_password is not None:
        expected_data['shared_link']['password'] = shared_link_password
    if shared_link_vanity_name is not None:
        expected_data['shared_link']['vanity_name'] = shared_link_vanity_name

    url = test_web_link.get_shared_link(
        access=shared_link_access,
        unshared_at=shared_link_unshared_at,
        password=shared_link_password,
        vanity_name=shared_link_vanity_name,
    )
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_data),
        headers=None,
        params=None,
    )
    assert url == test_url


def test_clear_unshared_at_for_shared_link(
        test_web_link,
        mock_box_session,
        test_url,
):
    expected_url = test_web_link.get_url()
    mock_box_session.put.return_value.json.return_value = {
        'type': test_web_link.object_type,
        'id': test_web_link.object_id,
        'shared_link': {
            'url': test_url,
            'unshared_at': None,
        },
    }
    expected_data = {'shared_link': {'unshared_at': None, }, }
    shared_link = test_web_link.get_shared_link(unshared_at=None)
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps(expected_data),
        headers=None,
        params=None,
    )
    assert shared_link is test_url


def test_remove_shared_link(test_web_link, mock_box_session):
    # pylint:disable=redefined-outer-name, protected-access
    expected_url = test_web_link.get_url()
    mock_box_session.put.return_value.json.return_value = {
        'type': test_web_link.object_type,
        'id': test_web_link.object_id,
        'shared_link': None,
    }
    removed = test_web_link.remove_shared_link()
    mock_box_session.put.assert_called_once_with(
        expected_url,
        data=json.dumps({'shared_link': None}),
        headers=None,
        params=None,
    )
    assert removed is True



import sqlalchemy
import uuid
from test.functional.mock_box.db_model import DbModel


class TokenModel(DbModel):
    """DB Model for Box tokens."""
    __tablename__ = 'box_token'

    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, autoincrement=True)  # pylint:disable=invalid-name
    owned_by_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_user.id'))
    authorized_application_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey('box_application.id'))
    expires_at = sqlalchemy.Column(sqlalchemy.DateTime)
    token_type = sqlalchemy.Column(sqlalchemy.Enum('access', 'refresh'), nullable=False)
    token = sqlalchemy.Column(sqlalchemy.String(32), default=lambda: uuid.uuid4().hex)



from typing import Any, Union

from boxsdk.object.enterprise import Enterprise
from boxsdk.object.folder import Folder
from boxsdk.object.metadata_template import MetadataTemplate
from boxsdk.object.retention_policy import RetentionPolicy
from boxsdk.object.retention_policy_assignment import RetentionPolicyAssignment


class BoxRetentionPolicyAssignment:

    def __init__(
            self,
            retention_policy: RetentionPolicy,
            assignee: Union['Folder', 'Enterprise', 'MetadataTemplate']
    ):
        self._retention_policy_assignment = retention_policy.assign(assignee)

    def __enter__(self) -> RetentionPolicyAssignment:
        return self._retention_policy_assignment

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._retention_policy_assignment.delete()



from typing import Any

from .client import Client
from ..util.log import setup_logging


class LoggingClient(Client):
    """
    Box client subclass which logs requests and responses.
    """
    def __init__(self, *args: Any, **kwargs: Any):
        setup_logging(None)
        super().__init__(*args, **kwargs)



from bottle import HTTPError, ServerAdapter
from functools import partial, wraps
from threading import Thread


RETRY_AFTER_HEADER = str('Retry-After')


def abort(code, message=None, headers=None):
    """
    Abort a request and send a response with the given code, and optional message and headers.
    :raises:
        :class:`HTTPError`
    """
    raise HTTPError(code, {'message': message}, headers=headers)


def retry_after(delay, code=429):
    """
    Abort a request and send a response, including a Retry-After header informing the client when a retry of
    the request will be accepted.
    """
    abort(code, headers={RETRY_AFTER_HEADER: delay})


def authorize(method):
    """Decorator for a method that requires authorization. Unauthorized requests will be aborted with a 401."""
    @wraps(method)
    def authorized_method(self, *args, **kwargs):
        skip_auth = kwargs.pop('skip_auth', False)
        if not skip_auth:
            self.check_authorization_header()
        return method(self, *args, **kwargs)
    return authorized_method


def rate_limit(method):
    """Decorator for a method that requires rate limiting. Too many requests will be aborted with a 429."""
    @wraps(method)
    def limited_method(self, *args, **kwargs):
        skip_limit = kwargs.pop('skip_limit', False)
        if not skip_limit:
            self.check_rate_limits()
        return method(self, *args, **kwargs)
    return limited_method


def _route(verb, app, route):
    """Helper decorator to apply methods to routes."""
    def routed_method(method):
        setattr(method, 'verb', verb)
        setattr(method, 'app', app)
        setattr(method, 'route', route)
        return method
    return routed_method


def log_request(method):
    """Decorator for a method to add its request to the request log."""
    @wraps(method)
    def logged_method(self, *args, **kwargs):
        skip_log = kwargs.pop('skip_log', False)
        if not skip_log:
            self.append_to_request_log()
        return method(self, *args, **kwargs)
    return logged_method


GET = partial(_route, 'GET')
POST = partial(_route, 'POST')
PUT = partial(_route, 'PUT')
DELETE = partial(_route, 'DELETE')
OPTIONS = partial(_route, 'OPTIONS')


class StoppableWSGIRefServer(ServerAdapter):
    """
    Subclass of built-in Bottle server adapter that allows the server to be stopped.
    This is important for testing, since we don't want to "serve forever".
    """
    def __init__(self, host='127.0.0.1', port=8080, **options):
        super().__init__(host, port, **options)
        self.srv = None
        self._thread = None

    def run(self, app):
        from wsgiref.simple_server import WSGIRequestHandler, WSGIServer
        from wsgiref.simple_server import make_server

        class FixedHandler(WSGIRequestHandler):
            def address_string(self):
                return self.client_address[0]

            parent = self

            def log_request(self, *args, **kw):
                if not self.parent.quiet:
                    return WSGIRequestHandler.log_request(self, *args, **kw)

        handler_cls = self.options.get('handler_class', FixedHandler)
        server_cls = self.options.get('server_class', WSGIServer)

        self.srv = make_server(self.host, self.port, app, server_cls, handler_cls)
        thread = Thread(target=self.srv.serve_forever)
        thread.daemon = True
        thread.start()
        self._thread = thread
        self.srv.wait = self.wait
        return self.srv

    def wait(self):
        self.srv.server_close()
        self._thread.join()



from typing import Any

from .base_object import BaseObject


class CollaborationAllowlistEntry(BaseObject):
    """Represents a allowlisted email domain for enterprise collaboration."""

    _item_type = 'collaboration_whitelist_entry'

    def get_url(self, *args: Any) -> str:
        """
        Gets the collaboration allowlist entries endpoint URL.

        :return:
            The collaboration allowlist entries endpoint URL.
        """
        return self._session.get_url('collaboration_whitelist_entries', self._object_id, *args)



from unittest.mock import Mock
import pytest
from requests import Session


@pytest.fixture(params=('GET', 'POST', 'PUT', 'DELETE', 'OPTIONS'))
def http_verb(request):
    return request.param


@pytest.fixture()
def mock_request(monkeypatch):
    mock_session_factory = Mock()
    mock_session_factory.return_value = session = Mock(Session(), request=Mock())
    monkeypatch.setattr('requests.Session', mock_session_factory)
    return session.request


@pytest.fixture(params=['generic_successful_request_response', 'server_error_request_response_502', 'server_error_request_response_503'])
def request_response(generic_successful_request_response, server_error_request_response_502, server_error_request_response_503, request):
    # pylint:disable=unused-argument
    return locals()[request.param]


@pytest.fixture
def make_request_with_request_response(mock_request, http_verb, test_url, access_token):

    def _make_request_with_request_response(network, request_response, **kwargs):
        mock_request.return_value = request_response
        response = network.request(http_verb, test_url, access_token, **kwargs)
        mock_request.assert_called_once_with(http_verb, test_url, **kwargs)
        assert response.access_token_used == access_token
        return response

    return _make_request_with_request_response


@pytest.fixture
def make_network_request(make_request_with_request_response, request_response):

    def _make_network_request(network, **kwargs):
        return make_request_with_request_response(network, request_response, **kwargs)

    return _make_network_request



import logging
import re
import sys

from collections.abc import Mapping
from typing import Union, IO, Optional

_no_logger = object()


class Logging:
    _has_setup = False
    KEYS_TO_SANITIZE = (
        'Authorization',
        'access_token',
        'refresh_token',
        'subject_token',
        'token',
        'client_id',
        'client_secret',
        'code',
        'shared_link',
        'download_url',
        'jwt_private_key',
        'jwt_private_key_passphrase',
        'password',
    )

    PROXY_KEYS_TO_SANITIZE = (
        'http',
        'https',
    )

    def setup_logging(self, stream_or_file=_no_logger, debug=False, name=None):
        if not self._has_setup:
            self._has_setup = True
            self._setup_logging(stream_or_file, debug, name)

    @staticmethod
    def _setup_logging(stream_or_file=_no_logger, debug=False, name=None):
        logger = logging.getLogger(name)
        if isinstance(stream_or_file, str):
            logger.addHandler(logging.FileHandler(stream_or_file, mode='a'))
        elif stream_or_file is not _no_logger:
            logger.addHandler(logging.StreamHandler(stream_or_file or sys.stdout))
        logger.setLevel(logging.DEBUG if debug else logging.INFO)

    @staticmethod
    def sanitize_value(value):
        return f'---{value[-4:]}'

    @staticmethod
    def sanitize_proxy_value(value: str) -> str:
        return re.sub(
            '^(.*://)(.*):(.*)(@.*)$',
            lambda repl: f'{repl.group(1)}{"---"}:{"---"}{repl.group(4)}',
            value
        )

    def sanitize_dictionary(self, dictionary: Mapping) -> Mapping:
        if not isinstance(dictionary, Mapping):
            return dictionary
        sanitized_dictionary = {}
        for key, value in dictionary.items():
            if key in self.KEYS_TO_SANITIZE and isinstance(value, str):
                sanitized_dictionary[key] = self.sanitize_value(value)
            elif key in self.PROXY_KEYS_TO_SANITIZE and isinstance(value, str):
                sanitized_dictionary[key] = self.sanitize_proxy_value(value)
            elif isinstance(value, Mapping):
                sanitized_dictionary[key] = self.sanitize_dictionary(value)
            else:
                sanitized_dictionary[key] = value
        return sanitized_dictionary


_logging = Logging()


def setup_logging(
        stream_or_file: Optional[Union[str, IO]] = _no_logger,
        debug: Optional[bool] = False,
        name: Optional[str] = None
) -> None:
    """
    Create a logger for communicating with the user or writing to log files.
    Sets the level to INFO or DEBUG, depending on the debug flag.

    If a stream or file is passed (or None is passed to stream_or_file), then
    a handler to that stream or file (stdout for None) is added to the logger.

    :param stream_or_file:
        The destination of the log messages. If None, stdout will be used.
    :param debug:
        Whether or not the logger will be at the DEBUG level (if False, the logger will be at the INFO level).
    :param name:
        The logging channel. If None, a root logger will be created.
    """
    _logging.setup_logging(stream_or_file, debug, name)


def sanitize_dictionary(dictionary: Mapping) -> dict:
    """
    Get a copy of a dictionary that has sensitive information redacted. Should be called on objects that will be
    logged or printed.

    :param dictionary:      Dictionary that may contain sensitive information.
    :return:                Copy of the dictionary with sensitive information redacted.
    """
    return _logging.sanitize_dictionary(dictionary)


logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = ['setup_logging', 'sanitize_dictionary']



from typing import Optional, Iterable, TYPE_CHECKING

from boxsdk.object.base_object import BaseObject
from boxsdk.pagination.limit_offset_based_object_collection import LimitOffsetBasedObjectCollection
from boxsdk.util.api_call_decorator import api_call

if TYPE_CHECKING:
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class Collection(BaseObject):
    """Box API endpoint for interacting with collections."""

    _item_type = 'collection'

    @api_call
    def get_items(
            self,
            limit: Optional[int] = None,
            offset: int = 0,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the items in a collection using limit-offset paging.

        :param limit:
            The maximum number of items to return per page. If not specified, then will use the server-side default.
        :param offset:
            The index at which to start returning items.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the items in the folder.
        """
        return LimitOffsetBasedObjectCollection(
            self.session,
            self.get_url('items'),
            limit=limit,
            fields=fields,
            offset=offset,
            return_full_pages=False,
        )



from itertools import product

import pytest

from boxsdk.object.base_object import BaseObject
from boxsdk.object.file import File
from boxsdk.object.folder import Folder
from boxsdk.object.group import Group
from boxsdk.object.user import User
from boxsdk.object.web_link import WebLink
from boxsdk.util.translator import Translator


_response_to_class_mapping = {}


@pytest.fixture
def bookmark_response(make_mock_box_request, mock_object_id):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'bookmark', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture
def box_note_response(make_mock_box_request, mock_object_id):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'boxnote', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture(autouse=True)
def translator_response(
        bookmark_response,
        box_note_response,
        mock_file_response,
        mock_folder_response,
        mock_group_response,
        mock_user_response,
        mock_web_link_response,
):
    # pylint:disable=redefined-outer-name
    _response_to_class_mapping['bookmark'] = (bookmark_response, BaseObject)
    _response_to_class_mapping['box_note'] = (box_note_response, BaseObject)
    _response_to_class_mapping['file'] = (mock_file_response, File)
    _response_to_class_mapping['folder'] = (mock_folder_response, Folder)
    _response_to_class_mapping['group'] = (mock_group_response, Group)
    _response_to_class_mapping['user'] = (mock_user_response, User)
    _response_to_class_mapping['web_link'] = (mock_web_link_response, WebLink)


@pytest.fixture(params=('scope', 'id', 'both'))
def metadata_template_response(request):
    response = {
        'type': 'metadata_template',
    }
    if request.param in ('scope', 'both'):
        response.update({
            'scope': 'enterprise',
            'templateKey': 'vContract',
        })
    if request.param in ('id', 'both'):
        response['id'] = '2f2e84e9-afdb-4e9d-b293-d6d1d932fc85'

    return response


@pytest.mark.parametrize('response_type', ['bookmark', 'box_note', 'file', 'folder', 'group', 'user'])
def test_translator_converts_response_to_correct_type(response_type):
    response, object_class = _response_to_class_mapping[response_type]
    assert type(Translator().get(response.json()['type']) == object_class)


def test_default_translator():
    assert isinstance(Translator._default_translator, Translator)   # pylint:disable=protected-access


@pytest.mark.parametrize(('extend_default_translator', 'new_child'), list(product([None, True], [False, True])))
def test_with_extend_default_translator(default_translator, extend_default_translator, new_child):
    item_type = 'foo'

    class Foo:
        pass

    kwargs = {}
    if extend_default_translator is not None:
        kwargs['extend_default_translator'] = extend_default_translator
    translator = Translator({item_type: Foo}, new_child=new_child, **kwargs)
    assert set(translator.items()).issuperset(default_translator.items())


@pytest.mark.parametrize('new_child', [False, True])
def test_without_extend_default_translator(new_child):
    item_type = 'foo'

    class Foo:
        pass

    mapping = {item_type: Foo}

    translator = Translator(mapping, extend_default_translator=False, new_child=new_child)
    assert translator == mapping


@pytest.mark.parametrize(('new_child', 'extend_default_translator'), list(product([None, True], [False, True])))
def test_with_new_child(new_child, extend_default_translator):
    item_type = 'foo'

    class Foo:
        pass

    mapping = {item_type: Foo}

    kwargs = {}
    if new_child is not None:
        kwargs['new_child'] = new_child

    translator = Translator(mapping, extend_default_translator=extend_default_translator, **kwargs)
    assert item_type in translator
    assert not translator.maps[0]
    with pytest.raises(KeyError):
        del translator[item_type]

    class Bar(Foo):
        pass

    translator.register(item_type, Bar)
    assert translator.get(item_type) is Bar
    assert mapping == {item_type: Foo}


@pytest.mark.parametrize('extend_default_translator', [False, True])
def test_without_new_child(extend_default_translator):
    item_type = 'foo'

    class Foo:
        pass

    mapping = {item_type: Foo}

    translator = Translator(mapping, new_child=False, extend_default_translator=extend_default_translator)
    assert item_type in translator
    assert translator.maps[0] is mapping

    class Bar(Foo):
        pass

    translator.register(item_type, Bar)
    assert translator.get(item_type) is Bar
    assert mapping == {item_type: Bar}

    del translator[item_type]
    assert not mapping


def test_translate(default_translator, mock_box_session):
    response_object = {
        'entries': [
            {
                'type': 'folder',
                'id': '11111',
                'name': 'Test Folder',
                'created_by': {
                    'type': 'user',
                    'id': '33333',
                    'name': 'Test User',
                },
            },
            {
                'type': 'file',
                'id': '22222',
                'name': 'Test File',
                'modified_by': {
                    'type': 'user',
                    'id': '33333',
                    'name': 'Test User',
                },
            },
        ],
    }

    results = default_translator.translate(mock_box_session, response_object=response_object)
    test_folder = results['entries'][0]
    test_file = results['entries'][1]

    assert isinstance(test_folder, Folder)
    assert isinstance(test_file, File)
    assert test_folder.object_id == '11111'
    assert test_folder.name == 'Test Folder'
    assert test_file.object_id == '22222'
    assert test_file.name == 'Test File'

    user_1 = test_folder.created_by
    user_2 = test_file.modified_by
    assert isinstance(user_1, User)
    assert isinstance(user_2, User)
    assert user_1 == user_2

    assert test_folder._session == mock_box_session  # pylint:disable=protected-access
    assert test_file._session == mock_box_session  # pylint:disable=protected-access
    assert user_1._session == mock_box_session  # pylint:disable=protected-access
    assert user_2._session == mock_box_session  # pylint:disable=protected-access

    # It should not modify the original
    assert isinstance(response_object['entries'][0], dict)
    assert isinstance(response_object['entries'][1], dict)


def test_translator_translates_metadata_template(default_translator, mock_box_session, metadata_template_response):
    metadata_template = default_translator.translate(mock_box_session, metadata_template_response)

    if metadata_template_response.get('id'):
        assert metadata_template.id == metadata_template_response['id']
        assert metadata_template.object_id == metadata_template_response['id']

    if metadata_template_response.get('scope') and metadata_template_response.get('templateKey'):
        assert metadata_template.scope == metadata_template_response['scope']
        assert metadata_template.template_key == metadata_template_response['templateKey']

    if metadata_template_response.get('id') is None:
        assert metadata_template.object_id is None



import json

from boxsdk.exception import BoxAPIException
from boxsdk.config import API
from boxsdk.object.terms_of_service import TermsOfService
from boxsdk.object.terms_of_service_user_status import TermsOfServiceUserStatus


def test_get(test_terms_of_service, mock_box_session):
    created_at = '2016-05-18T17:38:03-07:00'
    expected_url = f'{API.BASE_API_URL}/terms_of_services/{test_terms_of_service.object_id}'
    mock_box_session.get.return_value.json.return_value = {
        'type': 'terms_of_service',
        'id': test_terms_of_service.object_id,
        'created_at': created_at,
    }
    terms_of_service = test_terms_of_service.get()
    mock_box_session.get.assert_called_once_with(expected_url, headers=None, params=None)
    assert isinstance(terms_of_service, TermsOfService)
    assert terms_of_service.created_at == created_at
    assert terms_of_service.id == test_terms_of_service.object_id
    assert terms_of_service.type == test_terms_of_service.object_type


def test_update(test_terms_of_service, mock_box_session):
    new_text = 'This is new text'
    expected_url = f'{API.BASE_API_URL}/terms_of_services/{test_terms_of_service.object_id}'
    mock_box_session.put.return_value.json.return_value = {
        'type': 'terms_of_service',
        'id': test_terms_of_service.object_id,
        'text': new_text,
    }
    data = {
        'text': new_text,
    }
    updated_terms_of_service = test_terms_of_service.update_info(data=data)
    mock_box_session.put.assert_called_once_with(expected_url, data=json.dumps(data), headers=None, params=None)
    assert isinstance(updated_terms_of_service, TermsOfService)
    assert updated_terms_of_service.text == new_text
    assert updated_terms_of_service.id == test_terms_of_service.object_id
    assert updated_terms_of_service.type == test_terms_of_service.object_type


def test_accept_terms_of_service(test_terms_of_service, test_terms_of_service_user_status, mock_user, mock_box_session):
    # pylint:disable=redefined-outer-name
    expected_url = f"{API.BASE_API_URL}/terms_of_service_user_statuses"
    created_at = '2016-05-18T17:38:03-07:00'
    value = json.dumps({
        'tos': {
            'type': 'terms_of_service',
            'id': test_terms_of_service.object_id,
        },
        'is_accepted': True,
        'user': {
            'type': mock_user.object_type,
            'id': mock_user.object_id,
        },
    })
    mock_box_session.post.return_value.json.return_value = {
        'type': 'terms_of_service_user_status',
        'id': test_terms_of_service_user_status.object_id,
        'created_at': created_at,
    }
    new_terms_of_service_user_status = test_terms_of_service.accept(mock_user)
    mock_box_session.post.assert_called_once_with(expected_url, data=value)
    assert isinstance(new_terms_of_service_user_status, TermsOfServiceUserStatus)
    assert new_terms_of_service_user_status.id == test_terms_of_service_user_status.object_id
    assert new_terms_of_service_user_status.type == test_terms_of_service_user_status.object_type
    assert new_terms_of_service_user_status.created_at == created_at


def test_reject_terms_of_service(test_terms_of_service, test_terms_of_service_user_status, mock_user, mock_box_session):
    # pylint:disable=redefined-outer-name
    expected_url = f"{API.BASE_API_URL}/terms_of_service_user_statuses"
    created_at = '2016-05-18T17:38:03-07:00'
    value = json.dumps({
        'tos': {
            'type': 'terms_of_service',
            'id': test_terms_of_service.object_id,
        },
        'is_accepted': False,
        'user': {
            'type': mock_user.object_type,
            'id': mock_user.object_id,
        },
    })
    mock_box_session.post.return_value.json.return_value = {
        'type': 'terms_of_service_user_status',
        'id': test_terms_of_service_user_status.object_id,
        'created_at': created_at,
    }
    new_terms_of_service_user_status = test_terms_of_service.reject(mock_user)
    mock_box_session.post.assert_called_once_with(expected_url, data=value)
    assert isinstance(new_terms_of_service_user_status, TermsOfServiceUserStatus)
    assert new_terms_of_service_user_status.id == test_terms_of_service_user_status.object_id
    assert new_terms_of_service_user_status.type == test_terms_of_service_user_status.object_type
    assert new_terms_of_service_user_status.created_at == created_at


def test_get_user_status(test_terms_of_service, mock_user, test_terms_of_service_user_status, mock_box_session):
    expected_url = f"{API.BASE_API_URL}/terms_of_service_user_statuses"
    created_at = '2016-05-18T17:38:03-07:00'
    expected_params = {
        'tos_id': test_terms_of_service.object_id,
        'user_id': mock_user.object_id,
    }
    mock_user_status = {
        'type': 'terms_of_service_user_status',
        'id': test_terms_of_service_user_status.object_id,
        'created_at': created_at,
    }
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 2,
        'entries': [mock_user_status],
    }
    new_terms_of_service_user_status = test_terms_of_service.get_user_status(user=mock_user)
    mock_box_session.get.assert_called_once_with(expected_url, params=expected_params)
    assert isinstance(new_terms_of_service_user_status, TermsOfServiceUserStatus)
    assert new_terms_of_service_user_status.type == 'terms_of_service_user_status'
    assert new_terms_of_service_user_status.id == test_terms_of_service_user_status.object_id
    assert new_terms_of_service_user_status.created_at == created_at


def test_get_empty_user_status(test_terms_of_service, mock_user, mock_box_session):
    mock_box_session.get.return_value.json.return_value = {
        'total_count': 1,
        'entries': [],
    }
    new_terms_of_service_user_status = test_terms_of_service.get_user_status(user=mock_user)
    assert new_terms_of_service_user_status is None


def test_set_user_status(test_terms_of_service, mock_user, mock_box_session):
    expected_post_url = f"{API.BASE_API_URL}/terms_of_service_user_statuses"
    expected_put_url = f"{API.BASE_API_URL}/terms_of_service_user_statuses/{test_terms_of_service.object_id}"
    post_value = json.dumps({
        'tos': {
            'type': test_terms_of_service.object_type,
            'id': test_terms_of_service.object_id,
        },
        'is_accepted': True,
        'user': {
            'type': mock_user.object_type,
            'id': mock_user.object_id,
        },
    })
    put_value = json.dumps({
        'is_accepted': True
    })
    mock_box_session.post.side_effect = [BoxAPIException(status=409, message="Conflict")]
    mock_box_session.get.return_value.json.return_value = {
        'entries': [
            {
                'type': 'terms_of_service_user_status',
                'id': '42',
            },
        ],
    }
    mock_box_session.put.return_value.json.return_value = {
        'type': 'terms_of_service_user_status',
        'id': '12345',
        'tos': {
            'type': test_terms_of_service.object_type,
            'id': test_terms_of_service.object_id,
        },
        'is_accepted': True,
    }
    new_terms_of_service_user_status = test_terms_of_service.set_user_status(True, mock_user)
    mock_box_session.post.assert_called_once_with(expected_post_url, data=post_value)
    mock_box_session.put.assert_called_once_with(expected_put_url, data=put_value, headers=None, params=None)
    assert isinstance(new_terms_of_service_user_status, TermsOfServiceUserStatus)
    assert new_terms_of_service_user_status.type == 'terms_of_service_user_status'
    assert new_terms_of_service_user_status.id == '12345'
    assert new_terms_of_service_user_status.tos['type'] == test_terms_of_service.object_type
    assert new_terms_of_service_user_status.tos['id'] == test_terms_of_service.object_id
    assert new_terms_of_service_user_status.is_accepted is True



import copy

from collections.abc import Sequence
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from boxsdk.session.session import Session
    from boxsdk.util.translator import Translator
    from boxsdk.object.base_object import BaseObject


class Page(Sequence):
    """
    A sequence of BaseObjects that belong to a page returned from a paging api call.

    The Page makes available detailed response data for page requests.
    """
    _item_entries_key_name = "entries"

    def __init__(self, session: 'Session', response_object: dict):
        """
        :param session:
            The Box session used to make the request that generated the response.
        :param response_object:
            The parsed HTTP response from Box after requesting more pages.
        """
        super().__init__()
        self._session = session
        self._response_object = response_object

    @property
    def _translator(self) -> 'Translator':
        """
        The translator used for translating Box API JSON responses into `BaseAPIJSONObject` smart objects.
        """
        return self._session.translator

    @property
    def response_object(self) -> dict:
        """
        Return a copy of the response object for this Page.
        """
        return copy.deepcopy(self._response_object)

    def __getitem__(self, key: str) -> 'BaseObject':
        """
        Try to get the attribute from the API response object.

        :param key:
            The attribute to retrieve from the API response object.
        """
        item_json = self._response_object[self._item_entries_key_name][key]
        return self._translator.translate(self._session, item_json)

    def __len__(self) -> int:
        """
        Get the number of items in the page.
        """
        return len(self._response_object[self._item_entries_key_name])



from unittest.mock import Mock, patch
import uuid

from boxsdk.auth import redis_managed_oauth2


def test_redis_managed_oauth2_gets_tokens_from_redis_on_init(access_token, refresh_token):
    redis_server = Mock(redis_managed_oauth2.StrictRedis)
    redis_server.hvals.return_value = access_token, refresh_token
    unique_id = Mock()
    oauth2 = redis_managed_oauth2.RedisManagedOAuth2(
        client_id=None,
        client_secret=None,
        unique_id=unique_id,
        redis_server=redis_server,
    )
    redis_server.hvals.assert_called_once_with(unique_id)
    assert oauth2.unique_id is unique_id


def test_redis_managed_oauth2_gets_tokens_from_redis_during_refresh(access_token, refresh_token, new_access_token):
    new_refresh_token = uuid.uuid4().hex
    redis_server = Mock(redis_managed_oauth2.StrictRedis)
    redis_server.hvals.return_value = new_access_token, new_refresh_token
    unique_id = Mock()
    oauth2 = redis_managed_oauth2.RedisManagedOAuth2(
        access_token=access_token,
        refresh_token=refresh_token,
        client_id=None,
        client_secret=None,
        unique_id=unique_id,
        redis_server=redis_server,
    )
    assert oauth2.access_token == access_token
    redis_server.hvals.assert_not_called()

    assert oauth2.refresh('bogus_access_token') == (new_access_token, new_refresh_token)
    assert oauth2.access_token == new_access_token
    redis_server.hvals.assert_called_once_with(unique_id)


def test_redis_managed_oauth2_stores_tokens_to_redis_during_refresh(
        access_token,
        refresh_token,
        mock_box_session,
        successful_token_response,
):
    redis_server = Mock(redis_managed_oauth2.StrictRedis)
    redis_server.hvals.return_value = access_token, refresh_token
    unique_id = Mock()
    with patch.object(redis_managed_oauth2.RedisManagedOAuth2, '_update_current_tokens'):
        oauth2 = redis_managed_oauth2.RedisManagedOAuth2(
            client_id=None,
            client_secret=None,
            unique_id=unique_id,
            redis_server=redis_server,
            session=mock_box_session,
        )
    mock_box_session.request.return_value = successful_token_response
    oauth2.send_token_request({}, access_token=None, expect_refresh_token=True)
    redis_server.hmset.assert_called_once_with(unique_id, {'access': access_token, 'refresh': refresh_token})



import json
from typing import TYPE_CHECKING, Optional

from boxsdk.util.text_enum import TextEnum
from boxsdk.exception import BoxAPIException
from .base_object import BaseObject

if TYPE_CHECKING:
    from boxsdk.object.user import User
    from boxsdk.object.terms_of_service_user_status import TermsOfServiceUserStatus


class TermsOfServiceType(TextEnum):
    """An enum of possible terms of service types"""
    MANAGED = 'managed'
    EXTERNAL = 'external'


class TermsOfServiceStatus(TextEnum):
    """An enum of possible terms of service status"""
    ENABLED = 'enabled'
    DISABLED = 'disabled'


class TermsOfService(BaseObject):
    """Represents a Box terms of service."""

    _item_type = 'terms_of_service'

    def get_user_status(self, user: Optional['User'] = None) -> Optional['TermsOfServiceUserStatus']:
        """
        Get the terms of service user status.

        :param user:
            This is the user to get the status of the terms of service for. This defaults to current
            user.
        :returns:
            A :class:`TermsOfServiceUserStatus` object or None if no user status found.
        """
        url = self._session.get_url('terms_of_service_user_statuses')
        additional_params = {
            'tos_id': self.object_id,
        }
        if user is not None:
            additional_params['user_id'] = user.object_id
        box_response = self._session.get(url, params=additional_params)
        response_object = box_response.json()

        if not response_object['entries']:
            return None

        response = response_object['entries'][0]
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    def accept(self, user: Optional['User'] = None) -> 'TermsOfServiceUserStatus':
        """
        Accept a terms of service.

        :param user:
            The :class:`User` to assign the terms of service to.
        :returns:
            A newly created :class:`TermsOfServiceUserStatus` object
        """
        return self.set_user_status(is_accepted=True, user=user)

    def reject(self, user: Optional['User'] = None) -> 'TermsOfServiceUserStatus':
        """
        Reject a terms of service.

        :param user:
            The :class:`User` to assign the terms of service to.
        :returns:
            A newly created :class:`TermsOfServiceUserStatus` object
        """
        return self.set_user_status(is_accepted=False, user=user)

    def set_user_status(self, is_accepted: bool, user: Optional['User'] = None) -> 'TermsOfServiceUserStatus':
        """
        Create a terms of service user status.

        :param is_accepted:
            Indicates whether a use has accepted or rejected a terms of service.
        :param user:
            The :class:`User` to assign the terms of service to.
        :returns:
            A newly created :class:`TermsOfServiceUserStatus` object
        """
        url = self._session.get_url('terms_of_service_user_statuses')
        body = {
            'tos': {
                'type': self.object_type,
                'id': self.object_id,
            },
            'is_accepted': is_accepted,
        }
        if user is not None:
            body['user'] = {
                'type': user.object_type,
                'id': user.object_id,
            }
        translated_response = None
        try:
            box_response = self._session.post(url, data=json.dumps(body))
            response = box_response.json()
            translated_response = self.translator.translate(
                session=self._session,
                response_object=response,
            )
        except BoxAPIException as err:
            if err.status == 409:
                user_status = self.get_user_status(user)
                translated_response = user_status.update_info(data={'is_accepted': is_accepted})
        return translated_response



from unittest.mock import ANY, call
from boxsdk.config import API


def test_automatic_refresh(
        box_client,
        mock_box_network,
        generic_successful_response,
        successful_token_mock,
        unauthorized_response,
):
    mock_box_network.session.request.side_effect = [
        unauthorized_response,
        successful_token_mock,
        generic_successful_response,
    ]
    box_client.folder('0').get()
    assert mock_box_network.session.request.mock_calls == [
        call(
            'GET',
            f'{API.BASE_API_URL}/folders/0',
            headers=ANY,
            params=None,
            log_response_content=True,
        ),
        call(
            'POST',
            f'{API.OAUTH2_API_URL}/token',
            data=ANY,
            headers={'content-type': 'application/x-www-form-urlencoded', 'User-Agent': ANY, 'X-Box-UA': ANY},
            log_response_content=True,
        ),
        call(
            'GET',
            f'{API.BASE_API_URL}/folders/0',
            headers=ANY,
            params=None,
            log_response_content=True,
        ),
    ]



from threading import Event, Thread

import pytest
import requests

from boxsdk.object.folder import FolderSyncState
from boxsdk.object.event import Event as BoxEvent


@pytest.fixture
def box_events(box_client):
    return box_client.events()


@pytest.fixture
def move_target(box_client):
    return box_client.folder('0').create_subfolder('move target')


@pytest.fixture
def copy_target(box_client):
    return box_client.folder('0').create_subfolder('copy target')


@pytest.fixture
def assert_event(box_events):
    # pylint:disable=redefined-outer-name
    def helper(get_item, event_type, stream_position=0):
        item = get_item()
        events = box_events.get_events(stream_position=stream_position)['entries']
        assert len(events) == 1
        event = events[0]
        assert event['event_type'] == event_type
        assert event['source']['name'] == item.name
        assert event['source']['id'] == item.id
        assert isinstance(event, BoxEvent)

    return helper


def test_get_long_poll_url(box_client):
    options = box_client.events().get_long_poll_options()
    with pytest.raises(requests.Timeout):
        requests.get(options['url'], timeout=0.11)


def test_upload_causes_upload_event(uploaded_file, assert_event):
    # pylint:disable=redefined-outer-name
    assert_event(lambda: uploaded_file, 'ITEM_UPLOAD')


def test_create_folder_causes_create_event(created_subfolder, assert_event):
    # pylint:disable=redefined-outer-name
    assert_event(lambda: created_subfolder, 'ITEM_CREATE')


def test_move_file_causes_move_event(box_events, move_target, uploaded_file, assert_event):
    # pylint:disable=redefined-outer-name
    assert_event(lambda: uploaded_file.move(move_target), 'ITEM_MOVE', box_events.get_latest_stream_position())


def test_move_folder_causes_move_event(box_events, move_target, created_subfolder, assert_event):
    # pylint:disable=redefined-outer-name
    assert_event(lambda: created_subfolder.move(move_target), 'ITEM_MOVE', box_events.get_latest_stream_position())


def test_rename_file_causes_rename_event(box_events, uploaded_file, assert_event):
    # pylint:disable=redefined-outer-name
    updated_name = f'updated_{uploaded_file.name}'
    assert_event(lambda: uploaded_file.rename(updated_name), 'ITEM_RENAME', box_events.get_latest_stream_position())


def test_rename_folder_causes_rename_event(box_events, created_subfolder, assert_event):
    # pylint:disable=redefined-outer-name
    updated_name = f'updated_{created_subfolder.name}'
    assert_event(lambda: created_subfolder.rename(updated_name), 'ITEM_RENAME', box_events.get_latest_stream_position())


def test_copy_file_causes_copy_event(box_events, copy_target, uploaded_file, assert_event):
    # pylint:disable=redefined-outer-name
    assert_event(lambda: uploaded_file.copy(parent_folder=copy_target), 'ITEM_COPY', box_events.get_latest_stream_position())


def test_copy_folder_causes_copy_event(box_events, copy_target, created_subfolder, assert_event):
    # pylint:disable=redefined-outer-name
    assert_event(lambda: created_subfolder.copy(parent_folder=copy_target), 'ITEM_COPY', box_events.get_latest_stream_position())


@pytest.mark.xfail(reason='trash event has no source')
def test_delete_file_causes_trash_event(box_events, uploaded_file, assert_event):
    # pylint:disable=redefined-outer-name
    assert_event(uploaded_file.delete, 'ITEM_TRASH', box_events.get_latest_stream_position())


@pytest.mark.xfail(reason='trash event has no source')
def test_delete_folder_causes_trash_event(box_events, created_subfolder, assert_event):
    # pylint:disable=redefined-outer-name
    assert_event(created_subfolder.delete, 'ITEM_TRASH', box_events.get_latest_stream_position())


@pytest.mark.parametrize('sync_state,event_type', [
    (FolderSyncState.IS_SYNCED, 'ITEM_SYNC'),
    (FolderSyncState.NOT_SYNCED, 'ITEM_UNSYNC'),
])
def test_sync_folder_causes_sync_event(box_events, created_subfolder, assert_event, sync_state, event_type):
    # pylint:disable=redefined-outer-name
    assert_event(
        lambda: created_subfolder.update_sync_state(sync_state.value),
        event_type,
        box_events.get_latest_stream_position(),
    )


@pytest.fixture
def long_poll_generator(box_events, uploaded_file, request):
    # pylint:disable=redefined-outer-name
    generator = box_events.generate_events_with_long_polling(stream_position=0)

    def long_poll():
        for event in generator:
            long_poll_thread.events.append(event)
            long_poll_thread.event_ready.set()
            if long_poll_thread.should_stop_polling:
                return
            long_poll_thread.consumed_events.wait()
            long_poll_thread.consumed_events.clear()

    long_poll_thread = Thread(target=long_poll)

    long_poll_thread.should_stop_polling = False
    long_poll_thread.events = []
    long_poll_thread.event_ready = Event()
    long_poll_thread.consumed_events = Event()
    long_poll_thread.start()

    def fin():
        long_poll_thread.should_stop_polling = True
        uploaded_file.delete()
        long_poll_thread.event_ready.wait()
        generator.close()
        long_poll_thread.join()

    request.addfinalizer(fin)

    return long_poll_thread


def test_generate_events_with_long_polling(long_poll_generator, created_subfolder, uploaded_file):
    # pylint:disable=redefined-outer-name
    long_poll_generator.event_ready.wait()
    long_poll_generator.event_ready.clear()
    long_poll_generator.consumed_events.set()
    long_poll_generator.event_ready.wait()
    long_poll_generator.event_ready.clear()
    long_poll_generator.consumed_events.set()
    assert not long_poll_generator.event_ready.wait(timeout=0.01)

    assert len(long_poll_generator.events) == 2
    folder_event = next(e for e in long_poll_generator.events if e['source']['type'] == 'folder')
    file_event = next(e for e in long_poll_generator.events if e['source']['type'] == 'file')
    assert folder_event['event_type'] == 'ITEM_CREATE'
    assert file_event['event_type'] == 'ITEM_UPLOAD'
    assert folder_event['source']['id'] == created_subfolder.id
    assert file_event['source']['id'] == uploaded_file.id
    assert folder_event['source']['name'] == created_subfolder.name
    assert file_event['source']['name'] == uploaded_file.name

    del long_poll_generator.events[:]

    uploaded_file.rename(f'updated_{uploaded_file.name}')
    created_subfolder.rename(f'updated_{created_subfolder.name}')

    long_poll_generator.event_ready.wait()
    long_poll_generator.event_ready.clear()
    long_poll_generator.consumed_events.set()
    long_poll_generator.event_ready.wait()
    long_poll_generator.event_ready.clear()
    long_poll_generator.consumed_events.set()
    assert not long_poll_generator.event_ready.wait(timeout=0.01)

    assert len(long_poll_generator.events) == 2
    folder_event = next(e for e in long_poll_generator.events if e['source']['type'] == 'folder')
    file_event = next(e for e in long_poll_generator.events if e['source']['type'] == 'file')
    assert folder_event['event_type'] == 'ITEM_RENAME'
    assert file_event['event_type'] == 'ITEM_RENAME'
    assert folder_event['source']['id'] == created_subfolder.id
    assert file_event['source']['id'] == uploaded_file.id
    assert folder_event['source']['name'] == f'updated_{created_subfolder.name}'
    assert file_event['source']['name'] == f'updated_{uploaded_file.name}'



import copy
from typing import Any, Iterator, Iterable, Optional

from ..util.translator import Translator


class BaseAPIJSONObjectMeta(type):
    """
    Metaclass for Box API objects.

    Registers classes with the default translator, so that API responses can be
    translated to the correct type. This relies on the _item_type field, which
    must be defined in the class's namespace dict (and must be re-defined, in
    order to register a custom subclass), to match the 'type' field of the
    response json.  But the type-class mapping will only be registered if the
    module of the class is imported.

    For example, events returned from the API look like

    .. code-block:: json

        {'type': 'event', ...}

    so a class for that type could be created and registered with the default
    translator like this:

    .. code-block:: python

        class Event(BaseAPIJSONObject):
            _item_type = 'event'
            ...

    NOTE: The default translator registration functionality is a private
    implementation detail of the SDK, to make it easy to register the default
    API object classes with the default translator. For convenience and
    backwards-compatability, developers are allowed to re-define the _item_type
    field in their own custom subclasses in order to take advantage of this
    functionality, but are encouraged not to. Since this is a private
    implementation detail, it may change or be removed in any major or minor
    release. Additionally, it has the usual hazards of mutable global state.
    The supported and recommended ways for registering custom subclasses are:

    - Constructing a new :class:`Translator`, calling `Translator.register()`
      as necessary, and passing it to the :class:`BoxSession` constructor.
    - Calling `session.translator.register()` on an existing
      :class:`BoxSession`.
    - Calling `client.translator.register()` on an existing :class:`Client`.
    """

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        item_type = attrs.get('_item_type', None)
        if item_type is not None:
            Translator._default_translator.register(item_type, cls)   # pylint:disable=protected-access,no-member
            # Some types have - in them instead of _ in the API.
            if "-" in item_type:
                Translator._default_translator.register(item_type.replace("-", "_"), cls)   # pylint:disable=protected-access,no-member


class BaseAPIJSONObject(metaclass=BaseAPIJSONObjectMeta):
    """Base class containing basic logic shared between true REST objects and other objects (such as an Event)"""

    # :attr _item_type:
    #     (protected) The Box resource type that this class represents.
    #     For API object/resource classes this should equal the expected value
    #     of the 'type' field in API JSON responses. Otherwise, this should be
    #     `None`.
    #
    # NOTE: When defining a leaf class with an _item_type in this SDK, it's
    # also important to add the module name to __all__ in object/__init__.py,
    # so that it will be imported and registered with the default translator.
    _item_type: Optional[str] = None
    _untranslated_fields = ()

    def __init__(self, response_object: dict = None, **kwargs: Any):
        """
        :param response_object:
            A JSON object representing the object returned from a Box API request.
        """
        super().__init__(**kwargs)
        self._response_object = response_object or {}
        self.__dict__.update(self._response_object)

    def __getitem__(self, item: str) -> Any:
        """
        Try to get the attribute from the API response object.

        :param item:
            The attribute to retrieve from the API response object.
        """
        return self._response_object[item]

    def __contains__(self, item: str) -> bool:
        """
        Does the response object contains this item attribute?

        :param item:
            The attribute to check for in the API response object.

        """
        return item in self._response_object

    def __iter__(self) -> Iterator:
        """
        Get all of the keys of the API response object.
        """
        return iter(self._response_object)

    def __repr__(self) -> str:
        """Base class override. Return a human-readable representation using the Box ID or name of the object."""
        extra_description = f' - {self._description}' if self._description else ''
        description = f'<Box {self.__class__.__name__}{extra_description}>'
        return description

    @property
    def _description(self) -> str:
        """Return a description of the object if one exists."""
        return ""

    @property
    def object_type(self) -> str:
        """Return the Box type for the object.
        """
        return self._item_type

    @classmethod
    def untranslated_fields(cls) -> tuple:
        """
        The fields that should not be translated on this object.
        """
        return cls._untranslated_fields

    @classmethod
    def _untranslate(cls, value: Any) -> Iterable:
        """
        Untranslates a given object into a dictionary.

        :param value:
            The object to untranslate.
        :return:
            A dictionary containing the untranslated object.
        """
        if isinstance(value, BaseAPIJSONObject):
            return cls._untranslate(value._response_object)  # pylint:disable=protected-access
        if isinstance(value, dict):
            return {k: cls._untranslate(v) for (k, v) in value.items()}
        if isinstance(value, list):
            return [cls._untranslate(entry) for entry in value]

        return copy.copy(value)

    @property
    def response_object(self) -> Iterable:
        """
        Return the JSON object representing the object returned from a Box API request.
        """
        return self._untranslate(self)



import copy
import json

from unittest.mock import Mock, MagicMock
import pytest

from boxsdk.config import API, Client, Proxy
from boxsdk.network import default_network
from boxsdk.network.default_network import DefaultNetworkResponse, DefaultNetwork
from boxsdk.session.box_response import BoxResponse
from boxsdk.session.session import Session
from boxsdk.util.translator import Translator


@pytest.fixture(scope='function', autouse=True)
def original_default_translator():
    """A reference to the default translator, before the reference is changed by `default_translator` below."""
    return Translator._default_translator   # pylint:disable=protected-access


@pytest.yield_fixture(scope='function', autouse=True)
def default_translator(original_default_translator):
    """The default translator to use during the test.

    We don't want global state to mutate across tests. So before each test
    (because of autouse=True), we make a copy of the default translator, and
    assign this copy to Translator._default_translator. At the end of the test,
    we reset the reference.
    """
    try:
        translator = Translator(dict(copy.deepcopy(original_default_translator)), extend_default_translator=False, new_child=False)
        Translator._default_translator = translator   # pylint:disable=protected-access
        yield translator
    finally:
        Translator._default_translator = original_default_translator  # pylint:disable=protected-access


@pytest.fixture(scope='function')
def translator(default_translator):   # pylint:disable=unused-argument
    return Translator(extend_default_translator=True, new_child=True)


@pytest.fixture(scope='function')
def mock_box_session(translator):
    mock_session = MagicMock(Session)
    # pylint:disable=protected-access
    mock_session._api_config = mock_session.api_config = API()
    mock_session._client_config = mock_session.client_config = Client()
    mock_session._proxy_config = mock_session.proxy_config = Proxy()
    # pylint:enable=protected-access
    mock_session.get_url.side_effect = lambda *args, **kwargs: Session.get_url(mock_session, *args, **kwargs)
    mock_session.translator = translator
    return mock_session


@pytest.fixture()
def mock_box_session_2(translator):
    mock_session = MagicMock(Session)
    # pylint:disable=protected-access
    mock_session._api_config = API()
    mock_session._client_config = Client()
    mock_session._proxy_config = Proxy()
    # pylint:enable=protected-access
    mock_session.get_url.side_effect = lambda *args, **kwargs: Session.get_url(mock_session, *args, **kwargs)
    mock_session.translator = translator
    return mock_session


@pytest.fixture(scope='function')
def mock_network_layer():
    mock_network = Mock(DefaultNetwork)
    return mock_network


@pytest.fixture(autouse=True)
def prevent_tests_from_making_real_network_requests(monkeypatch):
    monkeypatch.delattr(default_network.requests.Session, 'request')


@pytest.fixture(scope='function')
def mock_user_response(mock_user_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'user', 'id': mock_user_id},
    )
    return mock_box_response


@pytest.fixture(scope='function')
def mock_group_response(mock_group_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'group', 'id': mock_group_id},
    )
    return mock_box_response


@pytest.fixture()
def make_mock_box_request():
    def inner(status_code=200, response_ok=True, response=None, content=None):
        mock_box_response = Mock(BoxResponse)
        mock_network_response = Mock(DefaultNetworkResponse)
        mock_box_response.network_response = mock_network_response
        mock_box_response.status_code = status_code
        mock_box_response.ok = response_ok
        if response is not None:
            mock_box_response.json.return_value = response
            mock_box_response.content = json.dumps(response).encode()
        else:
            mock_box_response.content = content
        return mock_box_response, mock_network_response
    return inner


@pytest.fixture(scope='function')
def mock_file_response(mock_object_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'file', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture(scope='function')
def mock_folder_response(mock_object_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'folder', 'id': mock_object_id},
    )
    return mock_box_response


@pytest.fixture(scope='function')
def mock_web_link_response(mock_object_id, make_mock_box_request):
    # pylint:disable=redefined-outer-name
    mock_box_response, _ = make_mock_box_request(
        response={'type': 'web_link', 'id': mock_object_id},
    )
    return mock_box_response



import json
from typing import Optional, Iterable, TYPE_CHECKING

from .base_endpoint import BaseEndpoint
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..util.api_call_decorator import api_call
from ..util.text_enum import TextEnum

if TYPE_CHECKING:
    from boxsdk.pagination.box_object_collection import BoxObjectCollection
    from boxsdk.object.collaboration_allowlist_entry import CollaborationAllowlistEntry
    from boxsdk.object.user import User
    from boxsdk.object.collaboration_allowlist_exempt_target import CollaborationAllowlistExemptTarget


class AllowlistDirection(TextEnum):
    """
    Used to determine the direction of the allowlist.
    """
    INBOUND = 'inbound'
    OUTBOUNT = 'outbound'
    BOTH = 'both'


class CollaborationAllowlist(BaseEndpoint):
    """Represents the allowlist of email domains that users in an enterprise may collaborate with."""

    @api_call
    def get_entries(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the entries in the collaboration allowlist using limit-offset paging.

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the entries in the allowlist.
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('collaboration_whitelist_entries'),
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def add_domain(self, domain: str, direction: str) -> 'CollaborationAllowlistEntry':
        """
        Add a new domain to the collaboration allowlist.

        :param domain:
            The email domain to add to the allowlist.
        :param direction:
            The direction in which collaboration should be allowed: 'inbound', 'outbound', or 'both'.
        :returns:
            The created allowlist entry for the domain.
        """
        url = self.get_url('collaboration_whitelist_entries')
        data = {
            'domain': domain,
            'direction': direction
        }
        response = self._session.post(url, data=json.dumps(data)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def get_exemptions(
            self,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """
        Get the list of exempted users who are not subject to the collaboration allowlist rules.

        :param limit:
            The maximum number of entries to return per page. If not specified, then will use the server-side default.
        :param marker:
            The paging marker to start paging from.
        :param fields:
            List of fields to request.
        :returns:
            An iterator of the exemptions to the allowlist.
        """
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('collaboration_whitelist_exempt_targets'),
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )

    @api_call
    def add_exemption(self, user: 'User') -> 'CollaborationAllowlistExemptTarget':
        """
        Exempt a user from the collaboration allowlist.

        :param user:
            The user to exempt from the allowlist.
        :returns:
            The created allowlist exemption.
        """
        url = self.get_url('collaboration_whitelist_exempt_targets')
        data = {
            'user': {
                'id': user.object_id  # pylint:disable=protected-access
            }
        }
        response = self._session.post(url, data=json.dumps(data)).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )



from typing import Any, Optional

from test.integration_new import util
from test.integration_new import CLIENT
from boxsdk.object.folder import Folder


class BoxTestFolder:

    def __init__(self, *, name: str = None, parent_folder: Optional[Folder] = None):
        if name is None:
            name = util.random_name()
        if parent_folder is None:
            parent_folder = CLIENT.root_folder()
        self._folder: Folder = parent_folder.create_subfolder(name=name)

    def __enter__(self) -> Folder:
        return self._folder

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        util.permanently_delete(self._folder)



from datetime import datetime, timedelta
import json
import random
import string
from io import IOBase
from typing import Optional, Union, Callable, TYPE_CHECKING, Any
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
import jwt

from boxsdk.auth.server_auth import ServerAuth

if TYPE_CHECKING:
    from boxsdk.network.network_interface import Network
    from boxsdk.object.user import User


class JWTAuth(ServerAuth):
    """
    Responsible for handling JWT Auth for Box Developer Edition. Can authenticate enterprise instances or app users.
    """
    _GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:jwt-bearer'

    def __init__(
            self,
            client_id: str,
            client_secret: str,
            enterprise_id: Optional[str],
            jwt_key_id: str,
            rsa_private_key_file_sys_path: Optional[str] = None,
            rsa_private_key_passphrase: Optional[Union[str, bytes]] = None,
            user: Optional[Union[str, 'User']] = None,
            store_tokens: Optional[Callable[[str, str], None]] = None,
            box_device_id: str = '0',
            box_device_name: str = '',
            access_token: str = None,
            session: Optional['Network'] = None,
            jwt_algorithm: str = 'RS256',
            rsa_private_key_data: Union[bytes, IOBase, RSAPrivateKey] = None,
            **kwargs
    ):
        """Extends baseclass method.

        Must pass exactly one of either `rsa_private_key_file_sys_path` or
        `rsa_private_key_data`.

        If both `enterprise_id` and `user` are non-`None`, the `user` takes
        precedence when `refresh()` is called. This can be overruled with a
        call to `authenticate_instance()`.

        :param client_id:
            Box API key used for identifying the application the user is authenticating with.
        :param client_secret:
            Box API secret used for making OAuth2 requests.
        :param enterprise_id:
            The ID of the Box Developer Edition enterprise.

            May be `None`, if the caller knows that it will not be
            authenticating as an enterprise instance / service account.

            If `user` is passed, this value is not used, unless
            `authenticate_instance()` is called to clear the user and
            authenticate as the enterprise instance.
        :param jwt_key_id:
            Key ID for the JWT assertion.
        :param rsa_private_key_file_sys_path:
            (optional) Path to an RSA private key file, used for signing the JWT assertion.
        :param rsa_private_key_passphrase:
            Passphrase used to unlock the private key. Do not pass a unicode string - this must be bytes.
        :param user:
            (optional) The user to authenticate, expressed as a Box User ID or
            as a :class:`User` instance.

            This value is not required. But if it is provided, then the user
            will be auto-authenticated at the time of the first API call or
            when calling `authenticate_user()` without any arguments.

            Should be `None` if the intention is to authenticate as the
            enterprise instance / service account. If both `enterprise_id` and
            `user` are non-`None`, the `user` takes precedense when `refresh()`
            is called.

            May be one of this application's created App User. Depending on the
            configured User Access Level, may also be any other App User or
            Managed User in the enterprise.

            <https://developer.box.com/en/guides/applications/>
            <https://developer.box.com/en/guides/authentication/select/>
        :param store_tokens:
            Optional callback to get access to tokens and store them. Callback method should take two
             paramaters - access_token: str and refresh_token: str - and it is not expected to return anything.
        :param box_device_id:
            Optional unique ID of this device. Used for applications that want to support device-pinning.
        :param box_device_name:
            Optional human-readable name for this device.
        :param access_token:
            Access token to use for auth until it expires.
        :param session:
            If specified, use it to make network requests. If not, the default session will be used.
        :param jwt_algorithm:
            Which algorithm to use for signing the JWT assertion. Must be one of 'RS256', 'RS384', 'RS512'.
        :param rsa_private_key_data:
            (optional) Contents of RSA private key, used for signing the JWT assertion. Do not pass a
            unicode string. Can pass a byte string, or a file-like object that returns bytes, or an
            already-loaded `RSAPrivateKey` object.
        """
        rsa_private_key = self._normalize_rsa_private_key(
            file_sys_path=rsa_private_key_file_sys_path,
            data=rsa_private_key_data,
            passphrase=rsa_private_key_passphrase,
        )
        del rsa_private_key_data
        del rsa_private_key_file_sys_path
        super().__init__(
            client_id=client_id,
            client_secret=client_secret,
            enterprise_id=enterprise_id,
            user=user,
            store_tokens=store_tokens,
            box_device_id=box_device_id,
            box_device_name=box_device_name,
            access_token=access_token,
            refresh_token=None,
            session=session,
            **kwargs
        )
        self._rsa_private_key = rsa_private_key
        self._jwt_algorithm = jwt_algorithm
        self._jwt_key_id = jwt_key_id

    def _fetch_access_token(self, subject_id: str, subject_type: str, now_time: Optional[datetime] = None) -> str:
        """
        Construct the claims used for JWT auth and send a request to get a JWT.
        Pass an enterprise ID to get an enterprise token (which can be used to provision/deprovision users),
        or a user ID to get a user token.

        :param subject_id:
            The enterprise ID or user ID to auth.
        :param subject_type:
            Either 'enterprise' or 'user'
        :param now_time:
            Optional. The current UTC time is needed in order to construct the expiration time of the JWT claim.
            If None, `datetime.utcnow()` will be used.
        :return:
            The access token for the enterprise or app user.
        """
        system_random = random.SystemRandom()
        jti_length = system_random.randint(16, 128)
        ascii_alphabet = string.ascii_letters + string.digits
        ascii_len = len(ascii_alphabet)
        jti = ''.join(ascii_alphabet[int(system_random.random() * ascii_len)] for _ in range(jti_length))
        if now_time is None:
            now_time = datetime.utcnow()
        now_plus_30 = now_time + timedelta(seconds=30)
        assertion = jwt.encode(
            {
                'iss': self._client_id,
                'sub': subject_id,
                'box_sub_type': subject_type,
                'aud': 'https://api.box.com/oauth2/token',
                'jti': jti,
                'exp': int((now_plus_30 - datetime(1970, 1, 1)).total_seconds()),
            },
            self._rsa_private_key,
            algorithm=self._jwt_algorithm,
            headers={
                'kid': self._jwt_key_id,
            },
        )
        data = {
            'grant_type': self._GRANT_TYPE,
            'client_id': self._client_id,
            'client_secret': self._client_secret,
            'assertion': assertion,
        }
        if self._box_device_id:
            data['box_device_id'] = self._box_device_id
        if self._box_device_name:
            data['box_device_name'] = self._box_device_name
        return self.send_token_request(data, access_token=None, expect_refresh_token=False)[0]

    @classmethod
    def _normalize_rsa_private_key(
            cls,
            file_sys_path: str,
            data: Union[bytes, IOBase, RSAPrivateKey],
            passphrase: Optional[Union[str, bytes]] = None
    ) -> Any:
        if len(list(filter(None, [file_sys_path, data]))) != 1:
            raise TypeError("must pass exactly one of either rsa_private_key_file_sys_path or rsa_private_key_data")
        if file_sys_path:
            with open(file_sys_path, 'rb') as key_file:
                data = key_file.read()
        if hasattr(data, 'read') and callable(data.read):
            data = data.read()
        if isinstance(data, str):
            try:
                data = data.encode('ascii')
            except UnicodeError as unicode_error:
                raise TypeError(
                    "rsa_private_key_data must contain binary data (bytes/str), not a text/unicode string"
                ) from unicode_error

        if isinstance(data, bytes):
            passphrase = cls._normalize_rsa_private_key_passphrase(passphrase)
            return serialization.load_pem_private_key(
                data,
                password=passphrase,
                backend=default_backend(),
            )
        if isinstance(data, RSAPrivateKey):
            return data
        raise TypeError(
            'rsa_private_key_data must be binary data (bytes/str), '
            'a file-like object with a read() method, '
            'or an instance of RSAPrivateKey, '
            f'but got {data.__class__.__name__!r}'
        )

    @staticmethod
    def _normalize_rsa_private_key_passphrase(passphrase: Any):
        if isinstance(passphrase, str):
            try:
                return passphrase.encode('ascii')
            except UnicodeError as unicode_error:
                raise TypeError(
                    "rsa_private_key_passphrase must contain binary data (bytes/str), not a text/unicode string"
                ) from unicode_error

        if not isinstance(passphrase, (bytes, type(None))):
            raise TypeError(
                f"rsa_private_key_passphrase must contain binary data (bytes/str), "
                f"got {passphrase.__class__.__name__!r}"
            )
        return passphrase

    @classmethod
    def from_settings_dictionary(cls, settings_dictionary: dict, **kwargs: Any) -> 'JWTAuth':
        """
        Create an auth instance as defined by the given settings dictionary.

        The dictionary should have the structure of the JSON file downloaded from the Box Developer Console.

        :param settings_dictionary:       Dictionary containing settings for configuring app auth.
        :return:                        Auth instance configured as specified by the config dictionary.
        """
        if 'boxAppSettings' not in settings_dictionary:
            raise ValueError('boxAppSettings not present in configuration')
        return cls(
            client_id=settings_dictionary['boxAppSettings']['clientID'],
            client_secret=settings_dictionary['boxAppSettings']['clientSecret'],
            enterprise_id=settings_dictionary.get('enterpriseID', None),
            jwt_key_id=settings_dictionary['boxAppSettings']['appAuth'].get('publicKeyID', None),
            rsa_private_key_data=settings_dictionary['boxAppSettings']['appAuth'].get('privateKey', None),
            rsa_private_key_passphrase=settings_dictionary['boxAppSettings']['appAuth'].get('passphrase', None),
            **kwargs
        )

    @classmethod
    def from_settings_file(cls, settings_file_sys_path: str, **kwargs: Any) -> 'JWTAuth':
        """
        Create an auth instance as defined by a JSON file downloaded from the Box Developer Console.
        See https://developer.box.com/en/guides/authentication/jwt/ for more information.

        :param settings_file_sys_path:    Path to the JSON file containing the configuration.
        :return:                        Auth instance configured as specified by the JSON file.
        """
        with open(settings_file_sys_path, encoding='utf-8') as config_file:
            config_dictionary = json.load(config_file)
            return cls.from_settings_dictionary(config_dictionary, **kwargs)



import base64
import hashlib
import hmac
from typing import Optional

from .base_object import BaseObject


class Webhook(BaseObject):
    """Represents a Box Webhook."""

    _item_type = 'webhook'

    @staticmethod
    def validate_message(
            body: bytes,
            headers: dict,
            primary_signature_key: str,
            secondary_signature_key: str = None
    ) -> bool:
        """
        Validates a `Webhook` message.

        :param body:
            The encoded webhook body.
        :param headers:
            The headers for the `Webhook` notification.
        :param primary_signature_key:
            The `Webhook` primary signature key for this application.
        :param secondary_signature_key:
            The `Webhook` secondary signature key for this application.
        :return:
            A `bool` indicating whether a webhook message was validated or not
        """

        primary_signature = _compute_signature(body, headers, primary_signature_key)
        if primary_signature is not None and hmac.compare_digest(primary_signature, headers.get('box-signature-primary')):
            return True

        if secondary_signature_key:
            secondary_signature = _compute_signature(body, headers, secondary_signature_key)
            if secondary_signature is not None and hmac.compare_digest(secondary_signature, headers.get('box-signature-secondary')):
                return True
            return False

        return False


def _compute_signature(body: bytes, headers: dict, signature_key: str) -> Optional[str]:
    """
    Computes the Hmac for the webhook notification given one signature key.

    :param body:
        The encoded webhook body.
    :param headers:
        The headers for the `Webhook` notification.
    :param signature_key:
        The `Webhook` signature key for this application.
    :return:
        An Hmac signature.
    """
    if signature_key is None:
        return None
    if headers.get('box-signature-version') != '1':
        return None
    if headers.get('box-signature-algorithm') != 'HmacSHA256':
        return None

    encoded_signature_key = signature_key.encode('utf-8')
    encoded_delivery_time_stamp = headers.get('box-delivery-timestamp').encode('utf-8')
    new_hmac = hmac.new(encoded_signature_key, digestmod=hashlib.sha256)
    new_hmac.update(body + encoded_delivery_time_stamp)
    signature = base64.b64encode(new_hmac.digest()).decode()
    return signature



import logging
from logging import getLogger
from pprint import pformat
import sys
import time
from typing import Any, Callable, Type, Optional

import requests
from requests import Response
from requests.structures import CaseInsensitiveDict
from urllib3 import HTTPResponse

from .network_interface import Network, NetworkResponse
from ..util.log import sanitize_dictionary


class DefaultNetwork(Network):
    """Implements the network interface using the requests library."""

    LOGGER_NAME = 'boxsdk.network'
    REQUEST_FORMAT = '\x1b[36m%(method)s %(url)s %(request_kwargs)s\x1b[0m'
    EXCEPTION_FORMAT = '\x1b[31mRequest "%(method)s %(url)s" failed with %(exc_type_name)s exception: %(exc_value)r\x1b[0m'

    def __init__(self):
        super().__init__()
        self._session = requests.Session()
        self._logger = getLogger(__name__)

    def request(self, method: str, url: str, access_token: str, **kwargs: Any) -> NetworkResponse:
        """Base class override.

        Make a network request using a requests.Session. Logs information about an API request and response.

        Also logs exceptions before re-raising them.

        The logging of the response is deferred to :class:`DefaultNetworkResponse`.
        See that class's docstring for more info.
        """
        log_response_content = kwargs.pop('log_response_content', True)
        self._log_request(method, url, **kwargs)
        # pylint:disable=abstract-class-instantiated
        try:
            return self.network_response_constructor(
                request_response=self._session.request(method, url, **kwargs),
                access_token_used=access_token,
                log_response_content=log_response_content
            )
        except Exception:
            self._log_exception(method, url, sys.exc_info())
            raise

    def retry_after(self, delay: float, request_method: Callable, *args: Any, **kwargs: Any) -> Any:
        """Base class override.
        Retry after sleeping for delay seconds.
        """
        time.sleep(delay)
        return request_method(*args, **kwargs)

    @property
    def network_response_constructor(self) -> Type['DefaultNetworkResponse']:
        """Baseclass override.

        A callable that accepts `request_response` and `access_token_used`
        keyword arguments for the :class:`DefaultNetworkResponse` constructor,
        and returns an instance of :class:`DefaultNetworkResponse`.
        """
        return DefaultNetworkResponse

    def _log_request(self, method: str, url: str, **kwargs: Any) -> None:
        """
        Logs information about the Box API request.

        :param method:
            The HTTP verb that should be used to make the request.
        :param url:
            The URL for the request.
        """
        if self._logger.isEnabledFor(logging.INFO):
            self._logger.info(
                self.REQUEST_FORMAT,
                {'method': method, 'url': url, 'request_kwargs': pformat(sanitize_dictionary(kwargs))},
            )

    def _log_exception(self, method: str, url: str, exc_info: Any) -> None:
        """Log information at WARNING level about the exception that was raised when trying to make the request.

        :param method:  The HTTP verb that was used to make the request.
        :param url:   The URL for the request.
        :param exc_info:  The exception info returned from `sys.exc_info()`.
        """
        if self._logger.isEnabledFor(logging.WARNING):
            exc_type, exc_value, _ = exc_info
            self._logger.warning(
                self.EXCEPTION_FORMAT,
                {'method': method, 'url': url, 'exc_type_name': exc_type.__name__, 'exc_value': exc_value},
            )


class DefaultNetworkResponse(NetworkResponse):
    """Implementation of the network interface using the requests library.

    :class:`requests.Response` has a few mutually-exclusive ways to read the
    content of the response:

        - With the `Response.raw` attribute, an `io.IOBase` instance returned
          from the `urllib3` library, that can be read once in chunks from
          beginning to end.
        - With `Response.iter_content()` and other iter_* generators, which
          also can only be read once and advance the `Response.raw` IO stream.
        - With the `Response.content` property (and other attributes such as
          `Response.text` and `Response.json()`), which reads and caches the
          remaining response content in memory. Can be accessed multiple times,
          but cannot be safely accessed if any of the previous mechanisms have
          been used at all. And if this property has already been accessed,
          then the other mechanisms will have been exhausted, and attempting to
          read from them will make it appear like the response content is
          empty.

    Any of these mechanisms may be used to read any response, regardless of
    whether `stream=True` or `stream=False` on the request.

    If the caller uses `Response.content`, then it is safe for
    :class:`DefaultNetwork` to also access it. But if the caller uses any of
    the streaming mechanisms, then it is not safe for :class:`DefaultNetwork`
    to ever read any of the content.

    The SDK logs only the response content of type JSON. Non-JSON responses, e.g.
    the content of the downloaded file should not be logged by the SDK at any time.
    In that case the response will be logged with a placeholder for the actual content.

    :param: `log_response_content` specifies wheather the response content should be logged or not.
    Its value is determined by `BoxRequest.expect_json_response` field, which is set to False,
    only if explicitly specified inside SDK method, e.g. self._session.get(url, expect_json_response=False).
    So the contenet of the response will be read by the logger only when the call inside an SDK method
    expects a JSON response. In this case we can be sure that the SDK method will read the content
    of the response using non-streaming mechanism and accessing content of the response by logger
    with `response.json()` or 'response.content` is safe.
    """

    _COMMON_RESPONSE_FORMAT = '"%(method)s %(url)s" %(status_code)s %(content_length)s\n%(headers)s\n%(content)s\n'
    SUCCESSFUL_RESPONSE_FORMAT = f'\x1b[32m{_COMMON_RESPONSE_FORMAT}\x1b[0m'
    ERROR_RESPONSE_FORMAT = f'\x1b[31m{_COMMON_RESPONSE_FORMAT}\x1b[0m'
    CONTENT_NOT_LOGGED = '<No content or content unavailable for logging>'

    def __init__(self, request_response: 'Response', access_token_used: str, log_response_content: bool = True):
        self._logger = getLogger(__name__)
        self._request_response = request_response
        self._access_token_used = access_token_used
        self._did_log = False
        self._json = None
        self.log(can_safely_log_content=log_response_content)

    def json(self) -> dict:
        """Base class override."""
        if self._json is None:
            self._json = self._request_response.json()
        return self._json

    @property
    def content(self) -> Optional[bytes]:
        """Base class override."""
        return self._request_response.content

    @property
    def status_code(self) -> int:
        """Base class override."""
        return self._request_response.status_code

    @property
    def ok(self) -> bool:
        """Base class override."""
        # pylint:disable=invalid-name
        return self._request_response.ok

    @property
    def headers(self) -> CaseInsensitiveDict:
        """Base class override."""
        return self._request_response.headers

    @property
    def response_as_stream(self) -> HTTPResponse:
        """Base class override."""
        return self._request_response.raw

    @property
    def access_token_used(self) -> str:
        """Base class override."""
        return self._access_token_used

    @property
    def request_response(self) -> Response:
        """
        The response returned from the Requests library.
        """
        return self._request_response

    def log(self, can_safely_log_content: bool = False) -> None:
        """Logs information about the Box API response.

        Will only execute once. Subsequent calls will be no-ops. This is
        partially because we only want to log responses once, and partially
        because this is necessary to prevent this method from infinite
        recursing with its use of the `content` property.

        :param can_safely_log_content:
            (optional) `True` if the caller is accessing the `content`
            property, `False` otherwise.

            As stated in the class docstring, it is unsafe for this logging
            method to access `content` unless the caller is also accessing it.

            Defaults to `False`.
        """
        if self._did_log:
            return
        self._did_log = True

        if self.ok:
            logger_method, logger_level, response_format = self._logger.info, logging.INFO, self.SUCCESSFUL_RESPONSE_FORMAT
        else:
            logger_method, logger_level, response_format = self._logger.warning, logging.WARNING, self.ERROR_RESPONSE_FORMAT

        if not self._logger.isEnabledFor(logger_level):
            return

        content_length = self.headers.get('Content-Length', None)
        content = self.CONTENT_NOT_LOGGED
        if can_safely_log_content:
            if content_length is None:
                content_length = str(len(self.content))

            # If possible, get the content as a JSON `dict`, that way
            # `pformat(content)` will return pretty-printed JSON.
            try:
                content = self.json()
            except ValueError:
                content = self.content
            content = pformat(sanitize_dictionary(content))
        if content_length is None:
            content_length = '?'
        logger_method(
            response_format,
            {
                'method': self.request_response.request.method,
                'url': self.request_response.request.url,
                'status_code': self.status_code,
                'content_length': content_length,
                'headers': pformat(self.headers),
                'content': content,
            },
        )

    def __repr__(self) -> str:
        return f'<Box Network Response ({self._request_response.request.method} {self._request_response.request.url} ' \
               f'{self.status_code})>'



import json
from typing import Any, Union, TYPE_CHECKING, Iterable, Optional
from .base_object import BaseObject
from ..pagination.marker_based_object_collection import MarkerBasedObjectCollection
from ..util.api_call_decorator import api_call

if TYPE_CHECKING:
    from boxsdk.object.metadata_template import MetadataTemplate
    from boxsdk.object.enterprise import Enterprise
    from boxsdk.object.folder import Folder
    from boxsdk.object.retention_policy_assignment import RetentionPolicyAssignment
    from boxsdk.pagination.box_object_collection import BoxObjectCollection


class RetentionPolicy(BaseObject):
    """Represents a Box retention policy."""
    _item_type = 'retention_policy'

    def get_url(self, *args: Any) -> str:
        """
        Returns the url for this retention policy.
        """
        return self._session.get_url('retention_policies', self._object_id, *args)

    @api_call
    def assign(
            self,
            assignee: Union['Folder', 'Enterprise', 'MetadataTemplate'],
            fields: Iterable[str] = None,
            start_date_field: Optional[str] = None,
    ) -> 'RetentionPolicyAssignment':
        """Assign a retention policy to a Box item

        :param assignee:
            The item to assign the retention policy on.
        :param fields:
            List of fields to request.
        :param start_date_field:
            The date the retention policy assignment begins.
            If the assigned_to type is metadata_template, this field can be a date field's metadata attribute key id.
        :returns:
            A :class:`RetentionPolicyAssignment` object.
        """
        url = self._session.get_url('retention_policy_assignments')
        body = {
            'policy_id': self.object_id,
            'assign_to': {
                'type': assignee.object_type,
                'id': assignee.object_id,
            }
        }
        params = {}
        if fields is not None:
            params['fields'] = ','.join(fields)
        if start_date_field is not None:
            body['start_date_field'] = start_date_field
        response = self._session.post(url, data=json.dumps(body), params=params).json()
        return self.translator.translate(
            session=self._session,
            response_object=response,
        )

    @api_call
    def assignments(
            self,
            assignment_type: Optional[str] = None,
            limit: Optional[int] = None,
            marker: Optional[str] = None,
            fields: Iterable[str] = None
    ) -> 'BoxObjectCollection':
        """Get the assignments for the retention policy.

        :param assignment_type:
            The type of retention policy assignment to retrieve. Can be set to 'folder', 'enterprise', or 'metadata_template'.
        :param limit:
            The maximum number of items to return.
        :param marker:
            The position marker at which to begin the response.
        :param fields:
            List of fields to request.
        :returns:
            An iterable of assignments in the retention policy.
        """
        additional_params = {}
        if assignment_type is not None:
            additional_params['assignment_type'] = assignment_type
        return MarkerBasedObjectCollection(
            session=self._session,
            url=self.get_url('assignments'),
            additional_params=additional_params,
            limit=limit,
            marker=marker,
            fields=fields,
            return_full_pages=False,
        )



import json
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import TYPE_CHECKING, Optional, Union, Any, Iterator as Iter
from boxsdk.pagination.page import Page

if TYPE_CHECKING:
    from boxsdk.session.session import Session
    from boxsdk.object.base_object import BaseObject


class BoxObjectCollection(Iterator, ABC):
    """
    An iterator that represents a collection of Box objects (BaseObject).

    A BoxObjectCollection instance contains everything it needs in order to retrieve and page through
    responses from Box API endpoints that return collections of Box objects.

    This class only has two public methods:

    1). next(), which returns either a Page (sequence of BaseObjects) or individual BaseObjects based on
    the constructor argument 'return_full_pages'.

    2). next_pointer(), which returns the pointer (either an offset or a marker, based on the endpoint) that
    will be used to retrieve the next page of Box objects. This pointer can be used when requesting new
    BoxObjectCollection instances that start off from a particular page, instead of from the very beginning.
    """
    _page_constructor = Page

    def __init__(
            self,
            session: 'Session',
            url: str,
            limit: Optional[int] = None,
            fields: Optional[Iter[str]] = None,
            additional_params: Optional[dict] = None,
            return_full_pages: bool = False,
            use_post: bool = False
    ):
        """
        :param session:
            The Box session used to make requests.
        :param url:
            The endpoint url to hit.
        :param limit:
            The number of entries for each page to return. The default, as well as the upper limit of this value,
            differs by endpoint. See https://developer.box.com/en/reference. If limit is set to None, then the default
            limit (returned by Box in the response) is used.
        :param fields:
            List of fields to request. If None, will return the default fields for the object.
        :param additional_params:
            Additional HTTP params to send in the request.
        :param return_full_pages:
            If True, then the returned iterator for this collection will return full pages of Box objects on each
            call to next(). If False, the iterator will return a single Box object on each next() call.
        :param use_post:
            If True, then the returned iterator will make POST requests with all the data in the body on each
            call to next().
            If False, the iterator will make GET requets with all the data as query params on each call to next().
        """
        super().__init__()
        self._session = session
        self._url = url
        self._limit = limit
        self._fields = fields
        self._additional_params = additional_params
        self._return_full_pages = return_full_pages
        self._has_retrieved_all_items = False
        self._all_items = None
        self._use_post = use_post

    def next(self) -> Union[Page, 'BaseObject']:
        """
        Returns either a Page (a Sequence of BaseObjects) or a BaseObject depending on self._return_full_pages.

        Invoking this method may make an API call to Box. Any exceptions that can occur while making requests
        may be raised in this method.
        """
        if self._all_items is None:
            self._all_items = self._items_generator()
        return next(self._all_items)

    __next__ = next

    def _items_generator(self) -> Union[Page, 'BaseObject']:
        while not self._has_retrieved_all_items:
            response_object = self._load_next_page()

            self._update_pointer_to_next_page(response_object)
            self._has_retrieved_all_items = not self._has_more_pages(response_object)
            page = self._page_constructor(self._session, response_object)

            if self._return_full_pages:
                yield page
            else:
                # It's possible for the Box API to return 0 items in a page, even if there are more items to be
                # retrieved on subsequent pages. When self._return_full_pages is True, then yielding a 0-item
                # page is fine because that's what the page returned.
                # But when we are iterating over individual items, and not pages, it's odd to yield a sequence of
                # Nones (for that page that had 0 items). So instead, we continue to request more pages until we
                # have Box objects to yield.
                if not page:
                    continue
                yield from page

    def _load_next_page(self) -> dict:
        """
        Request the next page of entries from Box. Raises any network-related exceptions, including BoxAPIException.
        Returns a parsed dictionary of the JSON response from Box
        """
        params = {}
        if self._limit is not None:
            params['limit'] = self._limit
        if self._additional_params:
            params.update(self._additional_params)
        params.update(self._next_page_pointer_params())
        if self._use_post:
            if self._fields:
                params['fields'] = self._fields
            box_response = self._session.post(self._url, data=json.dumps(params), headers={b'Content-Type': b'application/json'})
        else:
            if self._fields:
                params['fields'] = ','.join(self._fields)
            box_response = self._session.get(self._url, params=params)
        return box_response.json()

    @abstractmethod
    def _update_pointer_to_next_page(self, response_object: dict) -> None:
        """
        Update the internal pointer attribute of this class to what will be used to request the next page
        of Box objects.

        A "pointer" can either be a marker (for marker-based paging) or an offset (for limit-offset paging).

        :param response_object:
            The parsed HTTP response from Box after requesting more pages.
        """
        raise NotImplementedError

    @abstractmethod
    def _has_more_pages(self, response_object: dict) -> bool:
        """
        Are there more pages of entries to query Box for? This gets invoked after self._update_pointer_to_next_page().

        :param response_object:
            The parsed HTTP response from Box after requesting more pages.
        """
        raise NotImplementedError

    @abstractmethod
    def _next_page_pointer_params(self) -> dict:
        """
        The dict of HTTP params that specify which page of Box objects to retrieve.
        """
        raise NotImplementedError

    @abstractmethod
    def next_pointer(self) -> Any:
        """
        The pointer that will be used to request the next page of Box objects.

        For limit-offset based paging, this is an offset. For marker-based paging, this is a marker.

        The pointer only gets progressed upon successful page requests to Box.
        """
        raise NotImplementedError



from urllib import parse
from test.functional.mock_box.box import Box
from test.util.streamable_mock_open import streamable_mock_open
from unittest.mock import patch

import re
import pytest
import requests

from boxsdk.auth.oauth2 import OAuth2
from boxsdk.client import LoggingClient
from boxsdk.config import API
from boxsdk.network.default_network import DefaultNetwork
from boxsdk.session.session import Session, AuthorizedSession


@pytest.fixture(scope='session')
def network_layer():
    return DefaultNetwork()


@pytest.fixture()
def box_client(box_oauth, box_session):
    # pylint:disable=redefined-outer-name
    return LoggingClient(box_oauth, session=box_session)


@pytest.fixture
def unauthorized_session(network_layer):
    return Session(network_layer=network_layer)


@pytest.fixture
def box_session(box_oauth, network_layer):
    # pylint:disable=redefined-outer-name
    return AuthorizedSession(oauth=box_oauth, network_layer=network_layer)


@pytest.fixture()
def box_oauth(client_id, client_secret, user_login, unauthorized_session):
    # pylint:disable=redefined-outer-name
    oauth2 = OAuth2(client_id, client_secret, box_device_name='mock_box functional test', session=unauthorized_session)
    url, _ = oauth2.get_authorization_url('http://localhost')
    form = requests.get(url + '&box_login=' + user_login).content.decode('utf-8')
    form_action = re.search('action="([^"]*)"', form).group(1)
    auth_response = requests.post(form_action, allow_redirects=False, data={
        'login': user_login,
        'client_id': client_id,
        'client_secret': client_secret,
    })
    redirect_url = auth_response.headers['Location']
    query_string = parse.urlparse(redirect_url).query
    parsed_query_string_dict = parse.parse_qs(query_string)

    # Get the OAuth2 authorization code from `parsed_query_string_dict`
    # (NOTE: the values in the dictionary are lists of strings),
    # and if necessary decode it from a utf-8 encoded byte string to
    # a unicode string.
    auth_code = parsed_query_string_dict['code'][0]
    if isinstance(auth_code, bytes):
        auth_code = auth_code.decode('utf-8')

    oauth2.authenticate(auth_code)
    return oauth2


@pytest.fixture(scope='session')
def mock_box_server(request):
    box = Box()
    request.addfinalizer(box.shutdown)
    return box


@pytest.fixture(autouse=True)
def mock_box(mock_box_server, monkeypatch, client_id, client_secret, user_name, user_login):
    # pylint:disable=redefined-outer-name
    mock_box_server.reset_filesystem([(user_name, user_login)], [(client_id, client_secret, 0)])
    monkeypatch.setattr(API, 'BASE_API_URL', f'http://localhost:{Box.API_PORT}')
    monkeypatch.setattr(API, 'UPLOAD_URL', f'http://localhost:{Box.UPLOAD_PORT}')
    monkeypatch.setattr(API, 'OAUTH2_API_URL', f'http://localhost:{Box.OAUTH_API_PORT}')
    monkeypatch.setattr(API, 'OAUTH2_AUTHORIZE_URL', f'http://localhost:{Box.OAUTH_AUTHORIZE_PORT}')
    return mock_box_server


@pytest.fixture(params=[
    'foo.txt',
    'bar.docx',
    'foo.txt',
])
def file_name(request):
    return request.param


@pytest.fixture(params=[
    'some_folder',
    'Ѵȁćȁƭȉőń Ρȉćƭȕŕȅŝ',
])
def folder_name(request):
    return request.param


@pytest.fixture(scope='session')
def user_name():
    return 'User 1'


@pytest.fixture(scope='session')
def user_login():
    return 'user.1@example.com'


@pytest.fixture()
def uploaded_file(box_client, test_file_path, test_file_content, file_name):
    # pylint:disable=redefined-outer-name
    with patch('boxsdk.object.folder.open', streamable_mock_open(read_data=test_file_content), create=True):
        return box_client.folder('0').upload(test_file_path, file_name)


@pytest.fixture()
def created_subfolder(box_client, folder_name):
    # pylint:disable=redefined-outer-name
    return box_client.folder('0').create_subfolder(folder_name)



from unittest.mock import Mock

import datetime
import pytest

from pytest_lazyfixture import lazy_fixture

from boxsdk.util import datetime_formatter


@pytest.mark.parametrize(
    "valid_datetime_format",
    (
        "2035-03-04T10:14:24+14:00",
        "2035-03-04T10:14:24-04:00",
        lazy_fixture("mock_datetime_rfc3339_str"),
    ),
)
def test_leave_datetime_string_unchanged_when_rfc3339_formatted_str_provided(
    valid_datetime_format,
):
    formatted_str = datetime_formatter.normalize_date_to_rfc3339_format(
        valid_datetime_format
    )
    assert formatted_str == valid_datetime_format


@pytest.mark.parametrize(
    "other_datetime_format",
    (
        "2035-03-04T10:14:24.000+14:00",
        "2035-03-04 10:14:24.000+14:00",
        "2035/03/04 10:14:24.000+14:00",
        "2035/03/04T10:14:24+14:00",
        "2035/3/4T10:14:24+14:00",
        lazy_fixture('mock_timezone_aware_datetime_obj'),
    ),
)
def test_normalize_date_to_rfc3339_format_timezone_aware_datetime(
    other_datetime_format,
    mock_datetime_rfc3339_str,
):
    formatted_str = datetime_formatter.normalize_date_to_rfc3339_format(
        other_datetime_format
    )
    assert formatted_str == mock_datetime_rfc3339_str


@pytest.mark.parametrize(
    "timezone_naive_datetime",
    (
        "2035-03-04T10:14:24.000",
        "2035-03-04T10:14:24",
        lazy_fixture('mock_timezone_naive_datetime_obj')
    ),
)
def test_add_timezone_info_when_timezone_naive_datetime_provided(
    timezone_naive_datetime,
    mock_timezone_naive_datetime_obj,
):
    formatted_str = datetime_formatter.normalize_date_to_rfc3339_format(
        timezone_naive_datetime
    )

    local_timezone = datetime.datetime.now().tzinfo
    expected_datetime = mock_timezone_naive_datetime_obj.astimezone(
        tz=local_timezone
    ).isoformat(timespec="seconds")
    assert formatted_str == expected_datetime


def test_return_none_when_none_provided():
    assert datetime_formatter.normalize_date_to_rfc3339_format(None) is None


@pytest.mark.parametrize("inavlid_datetime_object", (Mock(),))
def test_throw_type_error_when_invalid_datetime_object_provided(
    inavlid_datetime_object,
):
    with pytest.raises(TypeError):
        datetime_formatter.normalize_date_to_rfc3339_format(inavlid_datetime_object)



from .dict_page import DictPage
from .limit_offset_based_object_collection import LimitOffsetBasedObjectCollection


class LimitOffsetBasedDictCollection(LimitOffsetBasedObjectCollection):
    """Represents a limit/offset-based collection of simple dicts, which are not translated into objects."""
    _page_constructor = DictPage



from typing import Any

from .base_object import BaseObject


class LegalHold(BaseObject):

    """Represents the legal hold policy for a file version"""
    _item_type = 'legal_hold'

    def get_url(self, *args: Any) -> str:
        return self._session.get_url('file_version_legal_holds', self._object_id, *args)



from typing import TYPE_CHECKING

from ..auth import DeveloperTokenAuth
from .client import Client

if TYPE_CHECKING:
    from boxsdk import OAuth2
    from boxsdk.session.session import Session


class DeveloperTokenClient(Client):
    """
    Box client subclass which authorizes with a developer token.
    """
    def __init__(self, oauth: 'OAuth2' = None, session: 'Session' = None):
        super().__init__(oauth=oauth or DeveloperTokenAuth(), session=session)



from functools import update_wrapper, wraps
from typing import Callable, Any

from ..object.cloneable import Cloneable


def api_call(method: Callable) -> Any:
    """
    Designates the decorated method as one that makes a Box API call.
    The decorated method can then accept a new keyword argument `extra_network_parameters`,
    a dictionary of key-value pairs to be passed to the network layer for API
    calls made by the method.

    The decorated method must belong to a subclass of `Cloneable` as using this
    decorator and then passing a `extra_network_parameters` parameter to the method will cause
    the object's clone method to be called.

    :param method:
        The method to decorate.
    :return:
        A wrapped method that can pass extra request data to the network layer.
    """
    return APICallWrapper(method)


class APICallWrapper:

    def __init__(self, func_that_makes_an_api_call: Callable):
        super().__init__()
        self._func_that_makes_an_api_call = func_that_makes_an_api_call
        self.__name__ = func_that_makes_an_api_call.__name__
        update_wrapper(self, func_that_makes_an_api_call)

    def __call__(self, cloneable_instance: 'Cloneable', *args: Any, **kwargs: Any) -> Any:
        return self.__get__(cloneable_instance, type(cloneable_instance))(*args, **kwargs)

    def __get__(self, _instance: Any, owner: Any) -> Any:
        # `APICallWrapper` is imitating a function. For native functions,
        # ```func.__get__(None, cls)``` always returns `func`.
        if _instance is None:
            return self

        if isinstance(owner, type) and not issubclass(owner, Cloneable):
            raise TypeError(
                f"descriptor {self.__name__!r} must be owned by a 'Cloneable' subclass, not {owner.__name__}"
            )
        expected_type = owner or Cloneable
        if not isinstance(_instance, expected_type):
            raise TypeError(
                f"descriptor {self.__name__!r} for {expected_type.__name__!r} objects "
                f"doesn't apply to {_instance.__class__.__name__!r} object"
            )

        @wraps(self._func_that_makes_an_api_call)
        def call(instance, *args, **kwargs):
            extra_network_parameters = kwargs.pop('extra_network_parameters', None)
            if extra_network_parameters:
                # If extra_network_parameters is specified, then clone the instance, and specify the parameters
                # as the defaults to be used.
                instance = instance.clone(instance.session.with_default_network_request_kwargs(extra_network_parameters))

            method = self._func_that_makes_an_api_call.__get__(instance, owner)
            return method(*args, **kwargs)

        # Since the caller passed a non-`None` instance to `__get__()`, they
        # want a bound method back, not an unbound function. Thus, we must bind
        # `call()` to `_instance` and then return that bound method.
        return call.__get__(_instance, owner)



from typing import Any, Tuple, Optional
from uuid import uuid4

from redis import StrictRedis, Redis
from redis.lock import Lock

from .oauth2 import OAuth2


class RedisManagedOAuth2Mixin(OAuth2):
    """
    Box SDK OAuth2 subclass.
    Allows for storing auth tokens in redis.
    """
    def __init__(self, unique_id: str = uuid4(), redis_server: Redis = None, *args: Any, **kwargs: Any):
        """
        :param unique_id:
            An identifier for this auth object. Auth instances which wish to share tokens must use the same ID.
        :param redis_server:
            An instance of a Redis server, configured to talk to Redis.
        """
        # pylint:disable=keyword-arg-before-vararg
        self._unique_id = unique_id
        self._redis_server = redis_server or StrictRedis()
        refresh_lock = Lock(redis=self._redis_server, name=f'{self._unique_id}_lock')
        super().__init__(*args, refresh_lock=refresh_lock, **kwargs)
        if self._access_token is None:
            self._get_and_update_current_tokens()

    @property
    def unique_id(self) -> str:
        """
        Get the unique ID used by this auth instance. Other instances can share tokens with this instance
        if they share the ID with this instance.
        """
        return self._unique_id

    def _get_tokens(self) -> Tuple[Optional[str], Optional[str]]:
        """
        Base class override.
        Gets the latest tokens from redis before returning them.
        """
        return self._redis_server.hvals(self._unique_id) or (None, None)

    def _store_tokens(self, access_token: str, refresh_token: str) -> None:
        """
        Base class override.
        Saves the refreshed tokens in redis.
        """
        super()._store_tokens(access_token, refresh_token)
        self._redis_server.hmset(self._unique_id, {'access': access_token, 'refresh': refresh_token})


class RedisManagedOAuth2(RedisManagedOAuth2Mixin):
    """
    OAuth2 subclass which uses Redis to manage tokens.
    """



import json
from unittest.mock import Mock, PropertyMock, ANY

import pytest

from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.pagination.marker_based_object_collection import MarkerBasedObjectCollection
from boxsdk.session.box_response import BoxResponse
from boxsdk.session.session import Session
from .box_object_collection_test_base import BoxObjectCollectionTestBase


class TestMarkerBasedObjectCollection(BoxObjectCollectionTestBase):
    """
    In order to conveniently mimic marker based paging for the purposes of this test, the markers ('next_marker' in
    the response) returned by the mock_session object is always going to be of the convention: "marker_i", where
    i is the starting 0-index based offset of the element.
    """

    NO_NEXT_MARKER = object()

    @staticmethod
    @pytest.fixture(params=['', None, NO_NEXT_MARKER])
    def next_marker_value_for_last_page(request):
        return request.param

    @pytest.fixture()
    def mock_items_response(self, entries, next_marker_value_for_last_page):
        """Baseclass override."""
        # pylint:disable=redefined-outer-name,arguments-differ
        def get_response(limit, marker):
            mock_box_response = Mock(BoxResponse)
            mock_network_response = Mock(DefaultNetworkResponse)
            mock_box_response.network_response = mock_network_response

            mock_json = {}
            # The marker string should be of format: "marker_i", where i is the offset. Parse that out.
            # If the marker is None, then begin paging from the start of the entries.
            offset = 0
            if marker is not None:
                offset = int(marker.split('_')[1])
            mock_json['entries'] = entries[offset:limit + offset]

            # A next_marker is only returned if there are more pages left.
            if (offset + limit) < len(entries):
                mock_json['next_marker'] = f'marker_{offset + limit}'
            elif next_marker_value_for_last_page is not self.NO_NEXT_MARKER:
                mock_json['next_marker'] = next_marker_value_for_last_page

            mock_box_response.json.return_value = mock_json
            mock_box_response.content = json.dumps(mock_json).encode()
            mock_box_response.status_code = 200
            mock_box_response.ok = True
            return mock_box_response
        return get_response

    @pytest.fixture()
    def mock_session(self, translator, mock_items_response):
        """Baseclass override."""
        mock_box_session = Mock(Session)
        type(mock_box_session).translator = PropertyMock(return_value=translator)

        def mock_items_side_effect(_, params):
            limit = params['limit']
            marker = params.get('marker', None)
            return mock_items_response(limit, marker)

        mock_box_session.get.side_effect = mock_items_side_effect
        return mock_box_session

    def _object_collection_instance(  # pylint:disable=arguments-differ
            self,
            session,
            limit,
            return_full_pages=False,
            starting_pointer=None,
            supports_limit_offset_paging=False
    ):
        """Baseclass override."""
        return MarkerBasedObjectCollection(
            session,
            '/some/endpoint',
            limit=limit,
            return_full_pages=return_full_pages,
            marker=starting_pointer,
            supports_limit_offset_paging=supports_limit_offset_paging,
        )

    @pytest.mark.parametrize('return_full_pages', (True, False))
    def test_object_collection_sets_next_pointer_correctly(self, mock_session, return_full_pages):
        object_collection = self._object_collection_instance(mock_session, limit=5, return_full_pages=return_full_pages)
        assert object_collection.next_pointer() is None
        object_collection.next()
        assert object_collection.next_pointer() == 'marker_5'

    @pytest.mark.parametrize('supports_limit_offset_paging', (True, False))
    def test_object_collection_specifies_marker_param(self, mock_session, supports_limit_offset_paging):
        object_collection = self._object_collection_instance(
            mock_session,
            limit=100,
            supports_limit_offset_paging=supports_limit_offset_paging
        )
        object_collection.next()

        # Assert
        expected_params = {'limit': 100}
        if supports_limit_offset_paging:
            expected_params['useMarker'] = True
        mock_session.get.assert_called_with(ANY, params=expected_params)



from logging import Logger
from io import BytesIO
from unittest.mock import patch, Mock
from test.util.streamable_mock_open import streamable_mock_open
import pytest

from boxsdk.network.default_network import DefaultNetworkResponse
from boxsdk.network import default_network


@pytest.fixture
def logger():
    return Mock(Logger)


@pytest.fixture(autouse=True)
def mocked_logger(logger):
    with patch.object(default_network, 'getLogger') as get_logger:
        get_logger.return_value = logger
        yield


def test_upload_then_update(box_client, test_file_path, test_file_content, update_file_content, file_name):
    with patch('boxsdk.object.folder.open', streamable_mock_open(read_data=test_file_content), create=True):
        file_object = box_client.folder('0').upload(test_file_path, file_name)
    assert file_object.name == file_name
    file_object_with_info = file_object.get()
    assert file_object_with_info.id == file_object.object_id
    assert file_object_with_info.name == file_name
    file_content = file_object.content()
    expected_file_content = test_file_content.encode('utf-8') if isinstance(test_file_content, str)\
        else test_file_content
    assert file_content == expected_file_content
    folder_items = box_client.folder('0').get_items(100)
    item = folder_items.next()
    item_count = 1
    for _ in folder_items:
        item_count += 1
    assert item_count == 1
    assert item.object_id == file_object.object_id
    assert item.name == file_object.name
    with patch('boxsdk.object.file.open', streamable_mock_open(read_data=update_file_content), create=True):
        updated_file_object = file_object.update_contents(test_file_path)
    assert updated_file_object.name == file_name
    file_object_with_info = updated_file_object.get()
    assert file_object_with_info.id == updated_file_object.object_id
    assert file_object_with_info.name == file_name
    file_content = updated_file_object.content()
    expected_file_content = update_file_content.encode('utf-8') if isinstance(update_file_content, str)\
        else update_file_content
    assert file_content == expected_file_content
    folder_items = box_client.folder('0').get_items(100)
    item = folder_items.next()
    item_count = 1
    for _ in folder_items:
        item_count += 1
    assert item_count == 1
    assert item.object_id == file_object.object_id
    assert item.name == file_object.name


def test_upload_then_download(box_client, test_file_path, test_file_content, file_name):
    with patch('boxsdk.object.folder.open', streamable_mock_open(read_data=test_file_content), create=True):
        file_object = box_client.folder('0').upload(test_file_path, file_name)
    writeable_stream = BytesIO()
    file_object.download_to(writeable_stream)
    expected_file_content = test_file_content.encode('utf-8') if isinstance(test_file_content, str)\
        else test_file_content
    assert writeable_stream.getvalue() == expected_file_content


def test_do_not_log_downloaded_file_content_stream(box_client, test_file_path, test_file_content, file_name, logger):
    with patch('boxsdk.object.folder.open', streamable_mock_open(read_data=test_file_content), create=True):
        file_object = box_client.folder('0').upload(test_file_path, file_name)
    writeable_stream = BytesIO()

    file_object.download_to(writeable_stream)

    expected_file_content = test_file_content.encode('utf-8') if isinstance(test_file_content, str) \
        else test_file_content
    assert writeable_stream.getvalue() == expected_file_content
    assert logger.info.call_args[0][1]['content'] == DefaultNetworkResponse.CONTENT_NOT_LOGGED


def test_do_not_log_content_of_downloaded_file(box_client, test_file_path, test_file_content, file_name, logger):
    with patch('boxsdk.object.folder.open', streamable_mock_open(read_data=test_file_content), create=True):
        file_object = box_client.folder('0').upload(test_file_path, file_name)

    file_content = file_object.content()

    expected_file_content = test_file_content.encode('utf-8') if isinstance(test_file_content, str) \
        else test_file_content
    assert file_content == expected_file_content
    assert logger.info.call_args[0][1]['content'] == DefaultNetworkResponse.CONTENT_NOT_LOGGED



# pylint:disable=no-value-for-parameter

from collections import OrderedDict
from itertools import chain

from enum import EnumMeta

__all__ = ['ExtendableEnumMeta']

from typing import Any


class ExtendableEnumMeta(EnumMeta):
    """A metaclass for enum hierarchies.

    This allows you to define hierarchies such as this:

        from box.util.compat import with_metaclass

        class EnumBase(with_metaclass(ExtendableEnumMeta, Enum)): pass

        class Enum1(EnumBase):
            A = 'A'

        class Enum2(EnumBase): pass

        class Enum2_1(Enum2):
            B = 'B'

        class Enum2_2(Enum2):
            C = 'C'

    and have all members be accessible on EnumBase (as well as have all members
    of Enum2_1 and Enum2_2 be available on Enum2) as if they had been defined
    there.

    Non-leaf classes still may not have members directly defined on them, as
    with standard enums.

    Most of the usual enum magic methods are extended: __contains__, __dir__,
    __getitem__, __getattr__, __iter__, __len__, and __reversed__. Only __new__
    is not extended; instead, a new method `lookup` is provided. The
    __members__ property is also extended.
    """

    def lookup(cls, value: Any) -> Any:
        """Custom value lookup, which does recursive lookups on subclasses.

        If this is a leaf enum class with defined members, this acts the same
        as __new__().

        But if this is a base class with no defined members of its own, it
        tries doing a value lookup on all its subclasses until it finds the
        value.

        NOTE: Because of the implementation details of Enum, this must be a new
        classmethod, and can't be implemented as __new__() [1].

        [1] <https://docs.python.org/3.5/library/enum.html#finer-points>

        :param value:
            The value to look up. Can be a value, or an enum instance.
        :raises:
            :class:`ValueError` if the value isn't found anywhere.
        """
        try:
            return cls(value)
        except ValueError as value_error:
            for subclass in cls.__subclasses__():
                try:
                    return subclass.lookup(value)
                except ValueError:
                    pass
            raise value_error

    @property
    def __members__(cls):
        members = OrderedDict(super().__members__)
        for subclass in cls.__subclasses__():
            members.update(subclass.__members__)
        return members

    def __contains__(cls, member):
        try:
            if super().__contains__(member):
                return True
        except TypeError:
            return False

        def in_(subclass):
            return member in subclass

        return any(map(in_, cls.__subclasses__()))

    def __dir__(cls):
        return list(set(super().__dir__()).union(*map(dir, cls.__subclasses__())))

    def __getitem__(cls, name):
        try:
            return super().__getitem__(name)
        except KeyError as key_error:
            for subclass in cls.__subclasses__():
                try:
                    return subclass[name]
                except KeyError:
                    pass
            raise key_error

    def __getattr__(cls, name):
        try:
            return super().__getattr__(name)
        except AttributeError as attribute_error:
            try:
                # If the super() call fails, don't call getattr() on all of the
                # subclasses. Instead, use __getitem__ to do this. This is
                # because we don't want to grab arbitrary attributes from
                # subclasses, only enum members. For enum members, __getattr__
                # and __getitem__ have the same behavior. And __getitem__ has
                # the advantage of never grabbing anything other than enum
                # members.
                return cls[name]  # pylint:disable=unsubscriptable-object
            except KeyError:
                pass

            raise attribute_error

    def __iter__(cls):
        return chain(super().__iter__(), chain.from_iterable(map(iter, cls.__subclasses__())))

    def __len__(cls):
        return super().__len__() + sum(map(len, cls.__subclasses__()))

    def __reversed__(cls):
        return reversed(list(cls))
